# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the table ordering functions."""

from typing import Any, cast
from unittest import mock
from urllib.parse import urlencode

from django.core.exceptions import ImproperlyConfigured
from django.db.models import F
from django.test import RequestFactory

from debusine.db.models import User
from debusine.test.django import TestCase
from debusine.web.views.table.ordering import (
    DateTimeOrdering,
    NumberOrdering,
    OrderBys,
    Ordering,
    Sort,
    StringOrdering,
    TableOrdering,
    TableOrderings,
    compute_sort_icon_current_next,
    order_by_tuple,
    render_sort_icon,
)


class OrderByTupleTests(TestCase):
    """Tests for :py:func:`order_by_tuple`."""

    def test_single(self) -> None:
        for val in (
            "name",
            F("name"),
            F("name").asc(),
            ["name"],
            [F("name")],
            [F("name").asc()],
        ):
            with self.subTest(val=repr(val)):
                self.assertEqual(
                    str(order_by_tuple(val)),
                    "(OrderBy(F(name), descending=False),)",
                )

    def test_multi(self) -> None:
        for val in (
            ["name1", "name2"],
            [F("name1"), F("name2")],
            [F("name1").asc(), F("name2").asc()],
            cast(OrderBys, ["name1", F("name2").asc()]),
            cast(OrderBys, [F("name1"), "name2"]),
        ):
            with self.subTest(val=repr(val)):
                self.assertEqual(
                    str(order_by_tuple(val)),
                    "(OrderBy(F(name1), descending=False),"
                    " OrderBy(F(name2), descending=False))",
                )


class OrderingTests(TestCase):
    """Tests for :py:class:`Ordering`."""

    def test_default_desc(self) -> None:
        """Test inferring descending with a single field."""
        for val in ("name", F("name"), F("name").asc()):
            with self.subTest(val=repr(val)):
                o = Ordering(val)
                self.assertEqual(list(o.options.keys()), ["asc", "desc"])
                self.assertEqual(
                    str(o.options["asc"].order_by),
                    "(OrderBy(F(name), descending=False),)",
                )
                self.assertEqual(
                    str(o.options["desc"].order_by),
                    "(OrderBy(F(name), descending=True),)",
                )

    def test_default_desc_multi(self) -> None:
        """Test inferring descending with a single field."""
        for val in (
            ("name1", "name2"),
            (F("name1"), F("name2")),
            (F("name1").asc(), F("name2").asc()),
        ):
            with self.subTest(val=repr(val)):
                o = Ordering(val)
                self.assertEqual(list(o.options.keys()), ["asc", "desc"])
                self.assertEqual(
                    str(o.options["asc"].order_by),
                    "(OrderBy(F(name1), descending=False),"
                    " OrderBy(F(name2), descending=False))",
                )
                self.assertEqual(
                    str(o.options["desc"].order_by),
                    "(OrderBy(F(name1), descending=True),"
                    " OrderBy(F(name2), descending=True))",
                )

    def test_asc_desc(self) -> None:
        """Test inferring descending with a single field."""
        o = Ordering("name1", "name2")
        self.assertEqual(list(o.options.keys()), ["asc", "desc"])
        self.assertEqual(
            str(o.options["asc"].order_by),
            "(OrderBy(F(name1), descending=False),)",
        )
        self.assertEqual(
            str(o.options["desc"].order_by),
            "(OrderBy(F(name2), descending=False),)",
        )

    def test_dash(self) -> None:
        """Test inferring descending with a single field."""
        o = Ordering("name", "-name")
        self.assertEqual(list(o.options.keys()), ["asc", "desc"])
        self.assertEqual(
            str(o.options["asc"].order_by),
            "(OrderBy(F(name), descending=False),)",
        )
        self.assertEqual(
            str(o.options["desc"].order_by),
            "(OrderBy(F(name), descending=True),)",
        )

    def test_kwargs(self) -> None:
        """Test passing fields as kwargs."""
        o = Ordering(asc="name", desc="-name")
        self.assertEqual(list(o.options.keys()), ["asc", "desc"])
        self.assertEqual(
            str(o.options["asc"].order_by),
            "(OrderBy(F(name), descending=False),)",
        )
        self.assertEqual(
            str(o.options["desc"].order_by),
            "(OrderBy(F(name), descending=True),)",
        )

    def test_arbitrary_incomplete(self) -> None:
        """Arbitrary orderings not fully specified: ImproperlyConfigured."""
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"ordering that isn't 'asc' or 'desc'"
            r" needs to be given as full 'Sort' instances",
        ):
            Ordering(foo="foo")

    def test_arbitrary(self) -> None:
        """Test arbitrary orderings, with default icon."""
        o = Ordering(foo=Sort("foo", "Label", "icon"))
        self.assertEqual(list(o.options.keys()), ["foo"])
        self.assertEqual(o.options["foo"].label, "Label")
        self.assertEqual(o.options["foo"].icon, "icon")
        self.assertEqual(
            str(o.options["foo"].order_by),
            "(OrderBy(F(foo), descending=False),)",
        )

    def test_arbitrary_default_icon(self) -> None:
        """Test arbitrary orderings, with default icon."""
        o = Ordering(foo=Sort("foo", "Label"))
        self.assertEqual(list(o.options.keys()), ["foo"])
        self.assertEqual(o.options["foo"].label, "Label")
        self.assertEqual(o.options["foo"].icon, "sort-alpha-down")
        self.assertEqual(
            str(o.options["foo"].order_by),
            "(OrderBy(F(foo), descending=False),)",
        )

    def test_infer(self) -> None:
        """Infer label and icon (defaults to string)."""
        o = Ordering("name")
        self.assertEqual(o.options["asc"].label, "Alphabetical order")
        self.assertEqual(o.options["asc"].icon, "sort-alpha-down")
        self.assertEqual(o.options["desc"].label, "Reverse alphabetical order")
        self.assertEqual(o.options["desc"].icon, "sort-alpha-up-alt")


class StringOrderingTests(TestCase):
    """Tests for :py:class:`StringOrdering`."""

    def test_infer(self) -> None:
        """Infer label and icon."""
        o = StringOrdering("field")
        self.assertEqual(o.options["asc"].label, "Alphabetical order")
        self.assertEqual(o.options["asc"].icon, "sort-alpha-down")
        self.assertEqual(o.options["desc"].label, "Reverse alphabetical order")
        self.assertEqual(o.options["desc"].icon, "sort-alpha-up-alt")


class NumberOrderingTests(TestCase):
    """Tests for :py:class:`NumberOrdering`."""

    def test_infer(self) -> None:
        """Infer label and icon."""
        o = NumberOrdering("field")
        self.assertEqual(o.options["asc"].label, "Lowest first")
        self.assertEqual(o.options["asc"].icon, "sort-numeric-down")
        self.assertEqual(o.options["desc"].label, "Highest first")
        self.assertEqual(o.options["desc"].icon, "sort-numeric-up-alt")


class DateTimeOrderingTests(TestCase):
    """Tests for :py:class:`DateTimeOrdering`."""

    def test_infer(self) -> None:
        """Infer label and icon."""
        o = DateTimeOrdering("field")
        self.assertEqual(o.options["asc"].label, "Oldest first")
        self.assertEqual(o.options["asc"].icon, "sort-numeric-down")
        self.assertEqual(o.options["desc"].label, "Newest first")
        self.assertEqual(o.options["desc"].icon, "sort-numeric-up-alt")


class SortIconTests(TestCase):
    """Tests for sort icon rendering."""

    def orderings(
        self, options: int, active: int | None = None
    ) -> list[TableOrdering[User]]:
        """Create ordering lists."""
        table = mock.Mock()
        table.prefix = ""
        table.dom_id = "tdom_id"
        table.request = RequestFactory().get("/")
        table_orderings: TableOrderings[User] = TableOrderings(table)

        orderings = [
            TableOrdering(
                table_orderings,
                Sort(f"foo{i}", f"Foo{i}", f"icon{i}"),
                "test",
                f"foo{i}",
                "Test",
                f"order=test.foo{i}",
            )
            for i in range(options)
        ]

        if active is not None:
            orderings[active].active = True

        return orderings

    def test_compute_sort_icon_current_next(self) -> None:
        """Test current,next selection."""
        for size, active, current, next_ in (
            (0, None, None, None),
            (1, None, None, 0),
            (1, 0, 0, None),
            (2, None, None, 0),
            (2, 0, 0, 1),
            (2, 1, 1, 0),
            (3, None, None, 0),
            (3, 0, 0, 1),
            (3, 1, 1, 2),
            (3, 2, 2, 0),
        ):
            with self.subTest(size=size, active=active):
                orderings = self.orderings(size, active)
                c, n = compute_sort_icon_current_next(orderings)
                if current is None:
                    self.assertIsNone(c)
                else:
                    self.assertIs(c, orderings[current])
                if next_ is None:
                    self.assertIsNone(n)
                else:
                    self.assertIs(n, orderings[next_])
                if orderings:
                    self.assertHTMLValid(render_sort_icon(orderings))

    def test_render_with_no_orderings(self) -> None:
        """If there are no orderings, render returns None."""
        self.assertEqual(render_sort_icon([]), "")

    def test_render_with_current_only(self) -> None:
        """If current is defined, the icon shows it."""
        orderings = self.orderings(1, 0)
        tree = self.assertHTMLValid(render_sort_icon(orderings))
        i = self.assertHasElement(tree, "body/span/i")
        self.assertEqual(i.get("class"), f"bi bi-{orderings[0].icon}")

    def test_render_with_next_only(self) -> None:
        """If next is defined, the icon is a link to it."""
        orderings = self.orderings(1, None)
        tree = self.assertHTMLValid(render_sort_icon(orderings))
        a = self.assertHasElement(tree, "body/a")
        self.assertEqual(a.get("href"), "?" + orderings[0].query_string)

    def test_render_with_both(self) -> None:
        """If next is defined, the icon is a link to it."""
        orderings = self.orderings(2, 0)
        tree = self.assertHTMLValid(render_sort_icon(orderings))
        a = self.assertHasElement(tree, "body/a")
        self.assertEqual(a.get("href"), "?" + orderings[1].query_string)
        i = self.assertHasElement(a, "i")
        self.assertEqual(i.get("class"), f"bi bi-{orderings[0].icon}")
        self.assertEqual(
            i.get("title"),
            f"{orderings[0].sort.label} → {orderings[1].sort.label}",
        )


class TableOrderingTests(TestCase):
    """Tests for :py:class:`TableOrdering`."""

    def _orderings(self) -> TableOrderings[User]:
        """Create a mock table."""
        return mock.Mock()

    def test_members(self) -> None:
        sort = Sort("sname", "Sort Label", "sicon")
        to: TableOrdering[User] = TableOrdering(
            self._orderings(), sort, "what", "how", "What", "order=what.how"
        )
        self.assertEqual(to.sort, sort)
        self.assertEqual(to.what, "what")
        self.assertEqual(to.how, "how")
        self.assertEqual(to.label_what, "What")
        self.assertEqual(to.query_string, "order=what.how")
        self.assertEqual(to.icon, "sicon")
        self.assertFalse(to.active)
        self.assertEqual(str(to), "what.how")
        self.assertEqual(repr(to), """TableOrdering("what.how")""")
        tm = self.assertHTMLValid(to.table_menu_item)
        self.assertTextContentEqual(tm, "What: Sort Label")
        cm = self.assertHTMLValid(to.column_menu_item)
        self.assertTextContentEqual(cm, "Sort Label")


class TableOrderingsTests(TestCase):
    """Tests for :py:class:`TableOrderings`."""

    def _orderings(
        self,
        prefix: str = "",
        default_order: str | None = None,
        order: str | None = None,
    ) -> TableOrderings[User]:
        """Create a mock table."""
        table = mock.Mock()
        table.prefix = prefix
        table.dom_id = "tdom_id"
        if default_order is not None:
            table.default_order = default_order

        data: dict[str, Any] = {}
        if order is not None:
            data[f"{prefix}order"] = order
        table.request = RequestFactory().get("/", data=data)
        table.make_link = lambda **kwargs: urlencode(kwargs)
        return TableOrderings(table)

    def test_init(self) -> None:
        orderings = self._orderings()
        self.assertEqual(orderings.field_name, "order")
        self.assertEqual(orderings.options, [])
        self.assertFalse(hasattr(orderings, "current"))

    def test_init_with_prefix(self) -> None:
        orderings = self._orderings(prefix="table-")
        self.assertEqual(orderings.field_name, "table-order")
        self.assertEqual(orderings.options, [])
        self.assertFalse(hasattr(orderings, "current"))

    def test_add(self) -> None:
        orderings = self._orderings()
        res = orderings.add("username", Ordering("username"))
        self.assertEqual(len(res), 2)

        to = res[0]
        self.assertIs(to.orderings, orderings)
        self.assertEqual(to.sort.label, "Alphabetical order")
        self.assertEqual(to.what, "username")
        self.assertEqual(to.how, "asc")
        self.assertEqual(to.label_what, "Username")
        self.assertEqual(to.query_string, "order=username.asc")
        self.assertFalse(to.active)

        to = res[1]
        self.assertIs(to.orderings, orderings)
        self.assertEqual(to.sort.label, "Reverse alphabetical order")
        self.assertEqual(to.what, "username")
        self.assertEqual(to.how, "desc")
        self.assertEqual(to.label_what, "Username")
        self.assertEqual(to.query_string, "order=username.desc")
        self.assertFalse(to.active)

        self.assertEqual(orderings.options, res)

    def test_parse_order(self) -> None:
        orderings = self._orderings()
        for order, exp_what, exp_how in (
            ("user", "user", None),
            ("-user", "user", "desc"),
            ("user.asc", "user", "asc"),
            ("user.desc", "user", "desc"),
            ("user.foo", "user", "foo"),
        ):
            with self.subTest(order=order):
                self.assertEqual(
                    orderings._parse_order(order), (exp_what, exp_how)
                )

    def test_find(self) -> None:
        orderings = self._orderings()
        asc, desc = orderings.add("username", Ordering("username"))
        for order, expected in (
            ("username", asc),
            ("-username", desc),
            ("username.asc", asc),
            ("username.desc", desc),
            ("username.foo", None),
            ("-username.asc", None),
            ("foo", None),
            ("foo.asc", None),
        ):
            self.assertIs(orderings._find(order), expected)

    def test_finalize(self) -> None:
        orderings = self._orderings(default_order="username")
        asc, desc = orderings.add("username", Ordering("username"))
        orderings.finalize()
        self.assertIs(orderings.current, asc)
        self.assertTrue(asc.active)
        self.assertFalse(desc.active)

    def test_finalize_wrong_default_ordering(self) -> None:
        orderings = self._orderings(default_order="foo")
        orderings.add("username", Ordering("username"))
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"default_order 'foo' does not match available ordering options",
        ):
            orderings.finalize()

    def test_finalize_selected_column(self) -> None:
        orderings = self._orderings(
            default_order="username", order="username.desc"
        )
        asc, desc = orderings.add("username", Ordering("username"))
        orderings.finalize()
        self.assertIs(orderings.current, desc)
        self.assertFalse(asc.active)
        self.assertTrue(desc.active)

    def test_finalize_empty(self) -> None:
        """Test finalize when there are no orderings."""
        orderings = self._orderings()
        orderings.finalize()
        self.assertEqual(orderings.options, [])
        self.assertFalse(hasattr(orderings, "current"))

    def test_order_queryset(self) -> None:
        user1 = self.playground.create_user("user1")
        user2 = self.playground.create_user("user2")
        user3 = self.playground.create_user("user3")

        for order, expected in (
            (None, [user1, user2, user3]),
            ("username", [user1, user2, user3]),
            ("-username", [user3, user2, user1]),
            ("username.asc", [user1, user2, user3]),
            ("username.desc", [user3, user2, user1]),
        ):
            with self.subTest(order=order):
                orderings = self._orderings(
                    default_order="username", order=order
                )
                orderings.add("username", Ordering("username"))
                orderings.finalize()
                self.assertQuerySetEqual(
                    orderings.order_queryset(
                        User.objects.filter(username__startswith="user")
                    ),
                    expected,
                )

    def test_what_menu_entries(self) -> None:
        orderings = self._orderings(
            default_order="username", order="username.desc"
        )
        orderings.add("username", Ordering("username"))
        orderings.add("email", Ordering("username"))
        orderings.finalize()
        entries = orderings.what_menu_entries

        self.assertEqual(len(entries), 2)
        tree = self.assertHTMLValid(entries[0])
        self.assertTextContentEqual(tree, "Username")
        self.assertElementHasClass(tree.body.a, "active")
        tree = self.assertHTMLValid(entries[1])
        self.assertTextContentEqual(tree, "Email")
        self.assertElementHasNoClass(tree.body.a, "active")

    def test_how_menu_entries(self) -> None:
        orderings = self._orderings(
            default_order="username", order="email.desc"
        )
        orderings.add("username", Ordering("username"))
        easc, edesc = orderings.add("email", Ordering("username"))
        orderings.finalize()
        entries = orderings.how_menu_entries

        self.assertEqual(len(entries), 2)
        tree = self.assertHTMLValid(entries[0])
        self.assertTextContentEqual(tree, easc.sort.label)
        self.assertElementHasNoClass(tree.body.a, "active")
        tree = self.assertHTMLValid(entries[1])
        self.assertTextContentEqual(tree, edesc.sort.label)
        self.assertElementHasClass(tree.body.a, "active")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Reverse-dependencies autopkgtest workflow."""

import json
import re
from collections.abc import Iterable
from dataclasses import dataclass
from functools import cached_property, partial
from itertools import chain
from typing import Any, Self, override

from debian.deb822 import PkgRelation
from django.contrib.postgres.expressions import ArraySubquery
from django.db.models import (
    BooleanField,
    F,
    Func,
    JSONField,
    OuterRef,
    Q,
    Value,
)
from django.db.models.fields.json import KT
from django.db.models.functions import JSONObject

from debusine.artifacts import SourcePackage
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianSourcePackage,
    DebianUpload,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Collection, CollectionItem
from debusine.db.models.lookups import LookupResultArtifact
from debusine.server.collections.lookup import artifact_ids
from debusine.server.workflows import workflow_utils
from debusine.server.workflows.models import (
    AutopkgtestWorkflowData,
    ReverseDependenciesAutopkgtestWorkflowData,
    ReverseDependenciesAutopkgtestWorkflowDynamicData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.regression_tracking import (
    RegressionTrackingWorkflow,
)
from debusine.tasks.models import (
    BackendType,
    InputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    LookupSingle,
)

_comma_sep_re = re.compile(r"\s*,\s*")


class NoBinaryNames(Exception):
    """Raised if a lookup result has no binary names."""


class ToJSONB(Func):
    """Convert any SQL value to `jsonb`."""

    function = "to_jsonb"
    output_field = JSONField()


class JSONBPathExists(Func):
    """Check whether a JSON path returns any item for the given JSON value."""

    function = "jsonb_path_exists"
    output_field = BooleanField()


@dataclass(frozen=True, kw_only=True)
class _SourcePackage:
    """A representation of a source package."""

    qa_suite: LookupSingle
    name: str
    version: str

    @classmethod
    def from_artifact_data(
        cls, qa_suite: LookupSingle, artifact_data: dict[str, Any]
    ) -> Self:
        """Construct a representation of a source package."""
        data = DebianSourcePackage(**artifact_data)
        return cls(qa_suite=qa_suite, name=data.name, version=data.version)

    def to_lookup(self) -> str:
        """Return a lookup for this source package."""
        # We could just use artifact IDs, but this is more informative.
        collection = (
            f"{self.qa_suite}@collections"
            if isinstance(self.qa_suite, int)
            else self.qa_suite
        )
        return f"{collection}/source-version:{self.name}_{self.version}"


@dataclass(frozen=True, kw_only=True)
class _BinaryPackage:
    """A representation of a binary package."""

    qa_suite: LookupSingle
    name: str
    version: str
    architecture: str

    @classmethod
    def from_artifact_data(
        cls, qa_suite: LookupSingle, artifact_data: dict[str, Any]
    ) -> Self:
        """Construct a representation of a binary package."""
        data = DebianBinaryPackage(**artifact_data)
        return cls(
            qa_suite=qa_suite,
            name=data.deb_fields["Package"],
            version=data.deb_fields["Version"],
            architecture=data.deb_fields["Architecture"],
        )

    def to_lookup(self) -> str:
        """Return a lookup for this binary package."""
        # We could just use artifact IDs, but this is more informative.
        collection = (
            f"{self.qa_suite}@collections"
            if isinstance(self.qa_suite, int)
            else self.qa_suite
        )
        return (
            f"{collection}/"
            f"binary-version:{self.name}_{self.version}_{self.architecture}"
        )


def _lookup_multiple_to_list(
    lookup: LookupMultiple,
) -> list[int | str | dict[str, Any]]:
    """
    Export a :py:class:`LookupMultiple` as a list.

    This is helpful when combining more than one such lookup.
    """
    exported = lookup.export()
    return exported if isinstance(exported, list) else [exported]


class ReverseDependenciesAutopkgtestWorkflow(
    RegressionTrackingWorkflow[
        ReverseDependenciesAutopkgtestWorkflowData,
        ReverseDependenciesAutopkgtestWorkflowDynamicData,
    ]
):
    """Run autopkgtest for all the reverse-deps of a package in a suite."""

    TASK_NAME = "reverse_dependencies_autopkgtest"

    @override
    def post_init(self) -> None:
        super().post_init()
        if self.data.backend == BackendType.AUTO:
            self.data.backend = BackendType.UNSHARE

    @override
    @cached_property
    def qa_suite(self) -> Collection:
        """The collection to search for reverse-dependencies."""
        qa_suite = super().qa_suite
        # This is optional for regression-tracking workflows in general, but
        # the data model for this workflow declares it as required.
        assert qa_suite is not None
        return qa_suite

    def get_binary_names(self) -> set[str]:
        """Return names of binary artifacts whose rdeps we want to test."""
        binary_names: set[str] = set()
        for result in self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        ):
            if result.artifact is not None:
                binary_data = result.artifact.create_data()
                match binary_data:
                    case DebianBinaryPackage():
                        binary_names.add(binary_data.deb_fields["Package"])
                    case DebianUpload():
                        binary_names.update(
                            binary_data.changes_fields["Binary"].split()
                        )
                    case _:
                        raise NoBinaryNames(
                            f"Artifact of category {result.artifact.category} "
                            f"has no binary packages"
                        )
            else:
                assert result.collection_item is not None
                binary_names.update(result.collection_item.data["binary_names"])
        return binary_names

    def get_reverse_dependencies(
        self,
    ) -> list[tuple[_SourcePackage, list[_BinaryPackage]]]:
        """
        Find source/binary packages that depend on the given package.

        :return: a list of tuples, each of which is a source package and its
          list of associated binary packages.
        """
        source_artifact = self.work_request.lookup_single(
            self.data.source_artifact, expect_type=LookupChildType.ARTIFACT
        ).artifact
        source_artifact = workflow_utils.locate_debian_source_package(
            "source_artifact", source_artifact
        )
        source_artifact_data = source_artifact.create_data()
        assert isinstance(source_artifact_data, DebianSourcePackage)

        suite_artifacts = partial(
            CollectionItem.objects.active().artifacts_in_collection,
            parent_collection=self.qa_suite,
        )

        binary_names = self.get_binary_names()
        # This regex is executed by PostgreSQL, not Python, although as it
        # happens everything that we need is common to both implementations.
        # See:
        # https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-POSIX-REGEXP
        depends_pattern = (
            rf"(?:^|\W)"
            rf"(?:{'|'.join(re.escape(name) for name in binary_names)})"
            rf"(?:\W|$)"
        )
        depends_pattern_json = json.dumps(depends_pattern)

        # Find possible source packages and their corresponding binaries
        # that may count as reverse-dependencies for which we should run
        # tests.  It's difficult to express the exact conditions in SQL, but
        # we try to at least reduce the set of rows to something more
        # manageable so that we can do detailed checks in Python.
        candidates = suite_artifacts(
            category=ArtifactCategory.SOURCE_PACKAGE,
        ).annotate(
            package=KT("data__package"),
            version=KT("data__version"),
            source_data=F("artifact__data"),
            binaries=ToJSONB(
                ArraySubquery(
                    suite_artifacts(
                        category=ArtifactCategory.BINARY_PACKAGE,
                    )
                    .annotate(
                        source_name=KT("data__srcpkg_name"),
                        source_version=KT("data__srcpkg_version"),
                        binary_name=KT("data__package"),
                        binary_version=KT("data__version"),
                        architecture=KT("data__architecture"),
                    )
                    .filter(
                        source_name=OuterRef("package"),
                        source_version=OuterRef("version"),
                    )
                    .values(
                        json=JSONObject(id="id", artifact_data="artifact__data")
                    )
                    .order_by("binary_name", "binary_version", "architecture")
                ),
            ),
            testsuite=KT("source_data__dsc_fields__Testsuite"),
            testsuite_triggers=KT(
                "source_data__dsc_fields__Testsuite-Triggers"
            ),
        )

        # The name is not equal to the name of the source package being
        # tested.
        candidates = candidates.exclude(package=source_artifact_data.name)
        # The name is not listed in packages_denylist.
        candidates = candidates.exclude(package__in=self.data.packages_denylist)
        # The name is listed in packages_allowlist (only tested if the field
        # is present and set).
        if self.data.packages_allowlist is not None:
            candidates = candidates.filter(
                package__in=self.data.packages_allowlist
            )
        # It has binary packages built from it.
        candidates = candidates.exclude(binaries=[])
        # Either any of its binary packages depend on any of the given
        # binary package names, or any of the given binary package names are
        # in its Testsuite-Triggers field.
        any_binary_matches_jsonpath = (
            # JSON path queries only have rather basic support in Django,
            # but they seem to be the easiest way to query for properties
            # that hold for any of the associated binary packages.
            f'$[*].artifact_data.deb_fields ? ('
            f'@."Pre-Depends" like_regex {depends_pattern_json} '
            f'|| @."Depends" like_regex {depends_pattern_json}'
            f')'
        )
        candidates = candidates.annotate(
            any_binary_matches=JSONBPathExists(
                "binaries", Value(any_binary_matches_jsonpath)
            )
        ).filter(
            Q(any_binary_matches=True)
            | Q(testsuite_triggers__regex=depends_pattern)
        )
        # Its Testsuite field has an item either equal to "autopkgtest" or
        # starting with "autopkgtest-pkg".
        candidates = candidates.filter(
            # Imprecise, since this is usually close enough; we'll
            # double-check it in Python.
            testsuite__contains="autopkgtest"
        )
        candidates = candidates.order_by("package", "version")

        revdeps: list[tuple[_SourcePackage, list[_BinaryPackage]]] = []
        for candidate in candidates.iterator():
            source_data = candidate.source_data  # type: ignore[attr-defined]
            binaries_data = candidate.binaries  # type: ignore[attr-defined]

            # Check that the Testsuite field is suitable.
            if not any(
                testsuite == "autopkgtest"
                or testsuite.startswith("autopkgtest-pkg")
                for testsuite in _comma_sep_re.split(
                    # The query above ensures that this exists.
                    source_data["dsc_fields"]["Testsuite"]
                )
            ):
                continue

            # Double-check Pre-Depends, Depends, and Testsuite-Triggers;
            # it's hard to do full package relationship field parsing in
            # SQL.
            all_dependencies: list[list[PkgRelation.ParsedRelation]] = []
            for binary_data in binaries_data:
                # The query above ensures that this exists.
                deb_fields = binary_data["artifact_data"]["deb_fields"]
                if "Pre-Depends" in deb_fields:
                    all_dependencies.extend(
                        PkgRelation.parse_relations(deb_fields["Pre-Depends"])
                    )
                if "Depends" in deb_fields:
                    all_dependencies.extend(
                        PkgRelation.parse_relations(deb_fields["Depends"])
                    )
            if "Testsuite-Triggers" in source_data["dsc_fields"]:
                all_dependencies.extend(
                    PkgRelation.parse_relations(
                        source_data["dsc_fields"]["Testsuite-Triggers"]
                    )
                )
            # TODO: Maybe we should do more accurate tests on
            # version/architecture restrictions etc.?  For now, this is
            # easy, but it may be a problem if it results in superfluous
            # autopkgtests that fail.
            all_dependency_names = {
                dep["name"] for dep in chain.from_iterable(all_dependencies)
            }
            if not all_dependency_names.intersection(binary_names):
                continue

            source = _SourcePackage.from_artifact_data(
                self.data.qa_suite, source_data
            )
            binaries = {
                binary_data["id"]: _BinaryPackage.from_artifact_data(
                    self.data.qa_suite, binary_data["artifact_data"]
                )
                for binary_data in binaries_data
            }

            # Pre-cache lookups so that the autopkgtest sub-workflows don't
            # have to do so much work.  The parameters must match the
            # lookups performed by autopkgtest sub-workflows.
            binary_collection_items = {
                item.id: item
                for item in CollectionItem.objects.filter(
                    id__in=binaries.keys()
                ).select_related("artifact")
            }
            for lookup, collection_item, expect_type in (
                (source.to_lookup(), candidate, LookupChildType.ARTIFACT),
                *(
                    (
                        binary.to_lookup(),
                        binary_collection_items[binary_id],
                        LookupChildType.ARTIFACT_OR_PROMISE,
                    )
                    for binary_id, binary in binaries.items()
                ),
            ):
                assert collection_item.artifact is not None
                self.work_request.lookup_single(
                    lookup,
                    expect_type=expect_type,
                    precache_result=LookupResultArtifact(
                        lookup=lookup,
                        parent_collection_lookup=lookup.split("/", 1)[0],
                        collection_item=collection_item,
                        artifact=collection_item.artifact,
                    ),
                )

            revdeps.append((source, list(binaries.values())))

        return revdeps

    def _populate_single(
        self, source: _SourcePackage, binaries: Iterable[_BinaryPackage]
    ) -> None:
        """Create an autopkgtest sub-workflow for a single source package."""
        context_lookup = LookupMultiple.model_validate(
            _lookup_multiple_to_list(self.data.binary_artifacts)
            + _lookup_multiple_to_list(self.data.context_artifacts)
        )

        wr, _ = self.work_request_ensure_child_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix=f"{self.data.prefix}{source.name}_{source.version}|",
                reference_prefix=(
                    f"{self.data.reference_prefix}"
                    f"{source.name}_{source.version}|"
                ),
                source_artifact=source.to_lookup(),
                binary_artifacts=LookupMultiple.model_validate(
                    sorted([binary.to_lookup() for binary in binaries])
                ),
                context_artifacts=context_lookup,
                qa_suite=self.data.qa_suite,
                reference_qa_results=self.data.reference_qa_results,
                enable_regression_tracking=self.data.enable_regression_tracking,
                update_qa_results=self.data.update_qa_results,
                vendor=self.data.vendor,
                codename=self.data.codename,
                backend=self.data.backend,
                architectures=self.data.architectures,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
                extra_repositories=self.data.extra_repositories,
                debug_level=self.data.debug_level,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=f"autopkgtests for {source.name}/{source.version}",
                step=f"autopkgtests-{source.name}/{source.version}",
            ),
            # TODO: Hardcoded for now; see
            # https://salsa.debian.org/freexian-team/debusine/-/issues/973.
            relative_priority=-5,
        )
        # The sub-workflow adds appropriate dependencies, and provides
        # results in the internal collection.
        self.orchestrate_child(wr)

    @override
    def populate(self) -> None:
        """Create autopkgtest sub-workflows for all reverse-dependencies."""
        for source, binaries in self.get_reverse_dependencies():
            self._populate_single(source, binaries)

    @override
    def compute_dynamic_data(
        self,
    ) -> ReverseDependenciesAutopkgtestWorkflowDynamicData:
        r"""
        Compute dynamic data for this workflow.

        :subject: package name of ``source_artifact``
        :parameter_summary: package name of ``source_artifact``\_"version"
        :source_artifact_id: id of the ``source_artifact``
        :binary_artifacts_ids: ids of the ``binary_artifacts``
        :context_artifacts_ids: ids of the ``context_artifacts``
        :qa_suite_id: id of the ``qa_suite``
        """
        source_artifact = workflow_utils.source_package(self)
        source_data = SourcePackage.create_data(source_artifact.data)

        binary_artifacts_result = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )
        context_artifacts_result = self.work_request.lookup_multiple(
            self.data.context_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )
        qa_suite_id = self.work_request.lookup_single(
            self.data.qa_suite, expect_type=LookupChildType.COLLECTION
        ).collection.id

        return ReverseDependenciesAutopkgtestWorkflowDynamicData(
            subject=source_data.name,
            parameter_summary=f"{source_data.name}_{source_data.version}",
            source_artifact_id=source_artifact.id,
            binary_artifacts_ids=artifact_ids(binary_artifacts_result),
            context_artifacts_ids=artifact_ids(context_artifacts_result),
            qa_suite_id=qa_suite_id,
        )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return input artifacts."""
        return [
            InputArtifactSingle(
                lookup=self.data.source_artifact,
                label="source_artifact",
                artifact_id=(
                    self.dynamic_data.source_artifact_id
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactMultiple.from_ids(
                lookup=self.data.binary_artifacts,
                label="binary_artifacts",
                artifact_ids=(
                    self.dynamic_data.binary_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactMultiple.from_ids(
                lookup=self.data.context_artifacts,
                label="context_artifacts",
                artifact_ids=(
                    self.dynamic_data.context_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
        ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for functions in debusine.signing.signing_utils."""

import warnings
from pathlib import Path

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase as DjangoTestCase
from nacl.public import PrivateKey

from debusine.signing.signing_utils import (
    SensitiveTemporaryDirectory,
    read_private_key,
)
from debusine.test import TestCase


class SensitiveTemporaryDirectoryTests(TestCase):
    """Unit tests for `SensitiveTemporaryDirectory`."""

    def test_cleanup_warnings_are_scary(self) -> None:
        """Warnings about lack of cleanup have a suitably alarming message."""
        temp_dir = SensitiveTemporaryDirectory()

        with warnings.catch_warnings(record=True) as warnings_log:
            warnings.simplefilter("always")
            del temp_dir

        self.assertEqual(len(warnings_log), 1)
        self.assertEqual(warnings_log[0].category, ResourceWarning)
        self.assertRegex(
            str(warnings_log[0].message),
            r"^Implicitly cleaning up <SensitiveTemporaryDirectory '.*'> "
            r"\(SENSITIVE\)",
        )


class ReadPrivateKeyTests(TestCase, DjangoTestCase):
    """Tests for `read_private_key`."""

    def create_private_key_file(
        self, contents: bytes, permissions: int = 0o600
    ) -> Path:
        """Create a temporary private key file."""
        file = self.create_temporary_file(contents=contents)
        file.chmod(permissions)
        return file

    def test_cannot_read_file(self) -> None:
        """Raise ImproperlyConfigured if we cannot read the file."""
        with self.assertRaisesRegex(ImproperlyConfigured, "Cannot read"):
            read_private_key(self.create_temporary_directory() / "nonexistent")

    def test_permissions_too_open(self) -> None:
        """Raise ImproperlyConfigured if file permissions are too open."""
        private_key_file = self.create_private_key_file(
            b"\0" * PrivateKey.SIZE, 0o640
        )

        with self.assertRaisesRegex(
            ImproperlyConfigured, "Permission too open"
        ):
            read_private_key(private_key_file)

    def test_wrong_size(self) -> None:
        """Raise ImproperlyConfigured if the file is the wrong size."""
        private_key_file = self.create_private_key_file(b"\0")

        with self.assertRaisesRegex(ImproperlyConfigured, "Cannot load key"):
            read_private_key(private_key_file)

    def test_good_key(self) -> None:
        """If the file contains a private key, return it."""
        private_key = PrivateKey.generate()
        private_key_file = self.create_private_key_file(bytes(private_key))

        self.assertEqual(read_private_key(private_key_file), private_key)



#!/usr/bin/env python3

# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Integration tests for Debusine's debian_pipeline workflow."""

import json
import socket
import subprocess
import unittest
from textwrap import dedent
from typing import override

import yaml

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    TaskTypes,
)
from debusine.tasks.models import BackendType
from utils.client import Client
from utils.common import Configuration
from utils.integration_test_helpers_mixin import IntegrationTestHelpersMixin
from utils.server import DebusineServer


class IntegrationWorkflowDebianPipelineTests(
    IntegrationTestHelpersMixin, unittest.TestCase
):
    """
    Integration test for the ``debian_pipeline`` workflow.

    These tests assume:
    - debusine-server is running
    - debusine-worker is running (connected to the server)
    - debusine-client is correctly configured
    """

    @override
    def setUp(self) -> None:
        super().setUp()
        # If debusine-server or nginx was launched just before this, the
        # server might not be available yet.
        self.assertTrue(
            DebusineServer.wait_for_server_ready(),
            f"debusine-server should be available (in "
            f"{Configuration.get_base_url()}) before the integration tests "
            f"are run",
        )

        self.architecture = subprocess.check_output(
            ["dpkg", "--print-architecture"], text=True
        ).strip()

    def add_artifact_to_suite(
        self, suite_name: str, artifact_id: int, **variables: str
    ) -> None:
        """Add a source or binary package artifact to a suite."""
        # TODO: We don't have a sensible command to do this yet.
        result = DebusineServer.execute_command(
            "shell",
            "-v",
            "0",
            "-c",
            dedent(
                f"""\
                from debusine.artifacts.models import *
                from debusine.db.models import *

                suite = Collection.objects.get(
                    category=CollectionCategory.SUITE, name={suite_name!r}
                )
                artifact = Artifact.objects.get(id={artifact_id})
                suite.manager.add_artifact(
                    artifact, user=system_user(), variables={variables!r}
                )
                """
            ),
        )
        self.assertEqual(result.returncode, 0)

    def find_workflow_children(
        self, workflow_id: int, **filters: str
    ) -> list[int]:
        """Find children of a given workflow, filtered by various criteria."""
        # TODO: We don't have a sensible command to do this yet.
        filter_expr = ", ".join(
            [f"{name}={value!r}" for name, value in filters.items()]
        )
        result = DebusineServer.execute_command(
            "shell",
            "-v",
            "0",
            "-c",
            dedent(
                f"""\
                import json

                from debusine.db.models import *

                workflow = WorkRequest.objects.get(id={workflow_id})
                children = workflow.children.filter({filter_expr})
                print(json.dumps([wr.id for wr in children]))
                """
            ),
        )
        self.assertEqual(result.returncode, 0)
        parsed_result = json.loads(result.stdout)
        assert isinstance(parsed_result, list)
        return parsed_result

    def verify_sbuild_artifacts(self, workflow_id: int) -> None:
        """Verify ``sbuild`` artifacts produced by a workflow."""
        [sbuild_workflow_id] = self.find_workflow_children(
            workflow_id, task_type=TaskTypes.WORKFLOW.value, task_name="sbuild"
        )
        [sbuild_task_id] = self.find_workflow_children(
            sbuild_workflow_id,
            task_type=TaskTypes.WORKER.value,
            task_name="sbuild",
        )
        sbuild_task = Client.execute_command(
            "work-request", "show", sbuild_task_id
        )

        debian_binary_package_artifacts = 0
        debian_upload_artifacts = 0
        debian_binary_packages_seen = set()
        for artifact in sbuild_task["artifacts"]:
            if artifact["category"] == ArtifactCategory.BINARY_PACKAGE:
                debian_binary_package_artifacts += 1

                self.assertEqual(artifact["data"]["srcpkg_name"], "hello")
                self.assertEqual(
                    artifact["data"]["deb_fields"]["Architecture"],
                    self.architecture,
                )

                package = artifact["data"]["deb_fields"]["Package"]
                debian_binary_packages_seen.add(package)
                version = artifact["data"]["deb_fields"]["Version"]
                self.assertEqual(
                    set(artifact["files"].keys()),
                    {f"{package}_{version}_{self.architecture}.deb"},
                )
            elif artifact["category"] == ArtifactCategory.UPLOAD:
                debian_upload_artifacts += 1

                self.assertEqual(artifact["data"]["type"], "dpkg")
                self.assertEqual(
                    artifact["data"]["changes_fields"]["Source"], "hello"
                )

                version = artifact["data"]["changes_fields"]["Version"]
                self.assertCountEqual(
                    artifact["files"].keys(),
                    [
                        f"hello_{version}_{self.architecture}.deb",
                        f"hello-dbgsym_{version}_{self.architecture}.deb",
                        f"hello_{version}_{self.architecture}.buildinfo",
                        f"hello_{version}_{self.architecture}.changes",
                    ],
                )

        self.assertEqual(debian_binary_package_artifacts, 2)
        self.assertEqual(debian_binary_packages_seen, {"hello", "hello-dbgsym"})
        self.assertEqual(debian_upload_artifacts, 1)

    def verify_autopkgtest_artifacts(self, qa_workflow_id: int) -> None:
        """Verify ``autopkgtest`` artifacts produced by a workflow."""
        [autopkgtest_workflow_id] = self.find_workflow_children(
            qa_workflow_id,
            task_type=TaskTypes.WORKFLOW.value,
            task_name="autopkgtest",
        )
        [autopkgtest_task_id] = self.find_workflow_children(
            autopkgtest_workflow_id,
            task_type=TaskTypes.WORKER.value,
            task_name="autopkgtest",
        )
        autopkgtest_task = Client.execute_command(
            "work-request", "show", autopkgtest_task_id
        )

        debian_autopkgtest_artifacts = 0
        for artifact in autopkgtest_task["artifacts"]:
            if artifact["category"] == ArtifactCategory.AUTOPKGTEST:
                debian_autopkgtest_artifacts += 1

                # Expected testinfo.json in the artifact
                self.assertIn("testinfo.json", artifact["files"].keys())

                # The log has a reasonable Content-Type
                self.assertEqual(
                    artifact["files"]["log"]["content_type"],
                    "text/plain; charset=us-ascii",
                )

                # Check some of the data contents
                self.assertIn("results", artifact["data"])
                self.assertIn("source_package", artifact["data"])

        self.assertEqual(debian_autopkgtest_artifacts, 1)

    def verify_lintian_artifacts(self, qa_workflow_id: int) -> None:
        """Verify ``lintian`` artifacts produced by a workflow."""
        [lintian_workflow_id] = self.find_workflow_children(
            qa_workflow_id,
            task_type=TaskTypes.WORKFLOW.value,
            task_name="lintian",
        )
        [lintian_task_id] = self.find_workflow_children(
            lintian_workflow_id,
            task_type=TaskTypes.WORKER.value,
            task_name="lintian",
        )
        lintian_task = Client.execute_command(
            "work-request", "show", lintian_task_id
        )

        debian_lintian_artifacts: list[str] = []

        for artifact in lintian_task["artifacts"]:
            if artifact["category"] == ArtifactCategory.LINTIAN:
                debian_lintian_artifacts.append(
                    artifact["data"]["architecture"]
                )

                # Expected analysis.json and lintian.txt
                self.assertCountEqual(
                    artifact["files"].keys(), ["analysis.json", "lintian.txt"]
                )

        self.assertCountEqual(
            debian_lintian_artifacts, ["source", self.architecture]
        )

    def verify_piuparts_artifacts(self, qa_workflow_id: int) -> None:
        """Verify ``piuparts`` artifacts produced by a workflow."""
        [piuparts_workflow_id] = self.find_workflow_children(
            qa_workflow_id,
            task_type=TaskTypes.WORKFLOW.value,
            task_name="piuparts",
        )
        [piuparts_task_id] = self.find_workflow_children(
            piuparts_workflow_id,
            task_type=TaskTypes.WORKER.value,
            task_name="piuparts",
        )
        piuparts_task = Client.execute_command(
            "work-request", "show", piuparts_task_id
        )

        debian_piuparts_artifacts: list[str] = []

        for artifact in piuparts_task["artifacts"]:
            if artifact["category"] == ArtifactCategory.PIUPARTS:
                debian_piuparts_artifacts.append(
                    artifact["data"]["architecture"]
                )

                # Expected piuparts.txt
                self.assertCountEqual(
                    artifact["files"].keys(), ["piuparts.txt"]
                )

        self.assertCountEqual(debian_piuparts_artifacts, [self.architecture])

    def verify_debdiff_artifacts(self, qa_workflow_id: int) -> None:
        """Verify ``debdiff`` artifacts produced by a workflow."""
        [debdiff_workflow_id] = self.find_workflow_children(
            qa_workflow_id,
            task_type=TaskTypes.WORKFLOW.value,
            task_name="debdiff",
        )
        [debdiff_source_task_id] = self.find_workflow_children(
            debdiff_workflow_id,
            task_type=TaskTypes.WORKER.value,
            task_name="debdiff",
            workflow_data_json__step="debdiff-source",
        )
        debdiff_source_task = Client.execute_command(
            "work-request", "show", debdiff_source_task_id
        )
        [debdiff_binary_task_id] = self.find_workflow_children(
            debdiff_workflow_id,
            task_type=TaskTypes.WORKER.value,
            task_name="debdiff",
            workflow_data_json__step=f"debdiff-binaries-{self.architecture}",
        )
        debdiff_binary_task = Client.execute_command(
            "work-request", "show", debdiff_binary_task_id
        )

        for task in (debdiff_source_task, debdiff_binary_task):
            debian_debdiff_artifacts = 0

            for artifact in task["artifacts"]:
                if artifact["category"] == ArtifactCategory.DEBDIFF:
                    debian_debdiff_artifacts += 1

                    # Expected debdiff.txt
                    self.assertCountEqual(
                        artifact["files"].keys(), ["debdiff.txt"]
                    )

            self.assertEqual(debian_debdiff_artifacts, 1)

    def verify_blhc_artifacts(self, qa_workflow_id: int) -> None:
        """Verify ``blhc`` artifacts produced by a workflow."""
        [blhc_workflow_id] = self.find_workflow_children(
            qa_workflow_id, task_type=TaskTypes.WORKFLOW.value, task_name="blhc"
        )
        [blhc_task_id] = self.find_workflow_children(
            blhc_workflow_id, task_type=TaskTypes.WORKER.value, task_name="blhc"
        )
        blhc_task = Client.execute_command("work-request", "show", blhc_task_id)

        debian_blhc_artifacts = 0

        for artifact in blhc_task["artifacts"]:
            if artifact["category"] == ArtifactCategory.BLHC:
                debian_blhc_artifacts += 1

                # Expected blhc.txt
                self.assertCountEqual(artifact["files"].keys(), ["blhc.txt"])

        # 1 artifact for each job
        self.assertEqual(debian_blhc_artifacts, 1)

    def test_debian_pipeline(self) -> None:
        """Create a ``debian_pipeline`` workflow."""
        # Create a suite to use as a reference (e.g. for debdiff), and add
        # source and binary packages for "hello" to it.  We'll also use this
        # source artifact to start our pipeline.
        suite_name = "trixie"
        self.create_suite(suite_name, architectures=[self.architecture])
        with self.apt_indexes(suite_name) as apt_path:
            source_artifact_id = self.create_artifact_source(apt_path, "hello")
            self.add_artifact_to_suite(
                suite_name,
                source_artifact_id,
                component="main",
                section="devel",
            )
            binary_artifact_id = self.create_artifact_binary(apt_path, "hello")
            self.add_artifact_to_suite(
                suite_name,
                binary_artifact_id,
                component="main",
                section="devel",
                priority="optional",
            )

        # Create a workflow template.
        template_name = "test-debian-pipeline"
        template_data = {
            "static_parameters": {
                "vendor": "debian",
                "codename": "trixie",
                "architectures": [self.architecture],
                "arch_all_build_architecture": self.architecture,
                "sbuild_backend": BackendType.UNSHARE,
                "qa_suite": f"{suite_name}@{CollectionCategory.SUITE}",
                "autopkgtest_backend": BackendType.UNSHARE,
                "lintian_backend": BackendType.UNSHARE,
                "piuparts_backend": BackendType.UNSHARE,
                "enable_debdiff": True,
            },
            "runtime_parameters": {
                "source_artifact": "any",
            },
        }
        workflow_template_result = Client.execute_command(
            "workflow-template",
            "create",
            "debian_pipeline",
            template_name,
            stdin=yaml.safe_dump(template_data),
        )
        assert workflow_template_result.returncode == 0

        # Start a workflow with the source artifact downloaded earlier.
        workflow_id = Client.execute_command(
            "workflow",
            "start",
            "--yaml",
            template_name,
            stdin=yaml.safe_dump({"source_artifact": source_artifact_id}),
        )["id"]
        assert Client.wait_for_work_request_completed(workflow_id, "success")

        # Verify artifacts produced by various parts of the workflow.
        self.verify_sbuild_artifacts(workflow_id)
        [qa_workflow_id] = self.find_workflow_children(
            workflow_id, task_type=TaskTypes.WORKFLOW.value, task_name="qa"
        )
        self.verify_autopkgtest_artifacts(qa_workflow_id)
        self.verify_lintian_artifacts(qa_workflow_id)
        self.verify_piuparts_artifacts(qa_workflow_id)
        self.verify_debdiff_artifacts(qa_workflow_id)
        self.verify_blhc_artifacts(qa_workflow_id)

    def test_publish(self) -> None:
        """Publish source and binary packages to a suite."""
        base_template_name = "base-publish"
        base_template_data = {
            "static_parameters": {
                "vendor": "debian",
                "codename": "trixie",
                "architectures": [self.architecture],
                "sbuild_backend": "unshare",
                "enable_autopkgtest": False,
                "enable_lintian": False,
                "enable_piuparts": False,
                "enable_blhc": False,
            },
            "runtime_parameters": {
                "source_artifact": "any",
            },
        }
        workflow_template_result = Client.execute_command(
            "workflow-template",
            "create",
            "debian_pipeline",
            base_template_name,
            stdin=yaml.safe_dump(base_template_data),
        )
        assert workflow_template_result.returncode == 0

        suite_name = "trixie-publish"
        self.create_suite(
            suite_name,
            architectures=[self.architecture],
            base_workflow_template=base_template_name,
        )
        self.add_suite_to_archive(suite_name)
        with self.apt_indexes("trixie") as apt_path:
            source_artifact_id = self.create_artifact_source(
                apt_path, "base-files"
            )

        workflow_id = Client.execute_command(
            "workflow",
            "start",
            "--yaml",
            f"publish-to-{suite_name}",
            stdin=yaml.safe_dump({"source_artifact": source_artifact_id}),
        )["id"]
        assert Client.wait_for_work_request_completed(workflow_id, "success")

        work_requests = Client.execute_command(
            "work-request", "list", "--yaml", "--limit", "1000"
        )
        update_suites_id = next(
            wr
            for wr in work_requests.parsed_stdout
            if wr["task_type"] == "Workflow"
            and wr["task_name"] == "update_suites"
        )["id"]
        assert Client.wait_for_work_request_completed(
            update_suites_id, "success"
        )

        archive_url = f"https://deb.{socket.getfqdn()}/debusine/System"
        with (
            self.apt_indexes(
                suite_name,
                url=archive_url,
                signed_by=self.get_archive_signing_key(),
            ) as apt_path,
            self._temporary_directory() as temp_path,
        ):
            subprocess.run(
                ["apt-get", "source", "--download-only", "base-files"],
                cwd=temp_path,
                env=self.make_apt_environment(apt_path),
                check=True,
            )
            subprocess.run(
                ["apt-get", "download", "base-files"],
                cwd=temp_path,
                env=self.make_apt_environment(apt_path),
                check=True,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine-admin command to manage assets.

Note: to make commands easier to be invoked from Ansible, we take care to make
them idempotent.
"""

import argparse
import sys
from collections.abc import Callable
from functools import partial
from typing import Any, IO, NoReturn, cast, override

from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser
from django.db import IntegrityError, transaction

from debusine.assets import AssetCategory
from debusine.db.context import context
from debusine.db.models import Asset, Group, Scope, Workspace
from debusine.db.models.assets import (
    AssetRole,
    AssetRoles,
    AssetUsage,
    AssetUsageRole,
    AssetUsageRoles,
)
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import (
    AssetUsageRoles as AssetUsageRolesPrinter,
)
from debusine.server.management.management_utils import (
    Assets,
    ResourceRoles,
    get_scope,
    get_workspace,
)


class Command(DebusineBaseCommand):
    """Command to manage assets."""

    help = "Manage assets"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action", required=True)

        create = subparsers.add_parser(
            "create", help=self.handle_create.__doc__
        )
        create.add_argument("category", help="Asset category")
        create.add_argument(
            "--workspace",
            metavar="scope/name",
            type=partial(get_workspace, require_scope=True),
            help="Create asset in this workspace",
        )
        create.add_argument(
            "--data",
            dest="data_file",
            type=argparse.FileType("r"),
            help=(
                "File path (or - for stdin) to read the data for the asset. "
                "YAML format. Defaults to stdin."
            ),
            default="-",
        )

        delete = subparsers.add_parser(
            "delete", help=self.handle_delete.__doc__
        )
        delete.add_argument(
            "asset",
            metavar="id",
            type=self.get_asset_by_id,
            help="Asset to delete",
        )
        delete.add_argument(
            '--yes', action='store_true', help='Skips confirmation of deletion'
        )

        list_ = subparsers.add_parser("list", help=self.handle_list.__doc__)
        list_.add_argument(
            "--yaml", action="store_true", help="Machine readable YAML output"
        )
        list_.add_argument(
            "--scope",
            type=get_scope,
            help="list only assets in this scope",
        )
        list_.add_argument(
            "--workspace",
            metavar="scope/name",
            type=partial(get_workspace, require_scope=True),
            help="list only assets in this workspace",
        )

        list_roles = subparsers.add_parser(
            "list_roles", help=self.handle_list_roles.__doc__
        )
        list_roles.add_argument(
            "asset",
            type=self.get_asset_by_id,
            metavar="id",
            help="Asset to inspect",
        )
        list_roles.add_argument(
            "--yaml", action="store_true", help="Machine readable YAML output"
        )

        grant_role = subparsers.add_parser(
            "grant_role", help=self.handle_grant_role.__doc__
        )
        grant_role.add_argument(
            "--workspace",
            metavar="scope/name",
            type=partial(get_workspace, require_scope=True),
            help="Workspace to apply permissions in",
        )
        grant_role.add_argument(
            "asset",
            metavar="id",
            type=self.get_asset_by_id,
            help="Asset to grant permissions on",
        )
        grant_role.add_argument("role_name", help="Name of the role to assign")
        grant_role.add_argument(
            "groups",
            nargs='+',
            metavar="group",
            help="Group(s) getting the role",
        )

        revoke_role = subparsers.add_parser(
            "revoke_role", help=self.handle_revoke_role.__doc__
        )
        revoke_role.add_argument(
            "--workspace",
            metavar="scope/name",
            type=partial(get_workspace, require_scope=True),
            help="Workspace to apply permissions in",
        )
        revoke_role.add_argument(
            "asset",
            type=self.get_asset_by_id,
            metavar="id",
            help="Asset to grant permissions on",
        )
        revoke_role.add_argument("role_name", help="Name of the role to revoke")
        revoke_role.add_argument(
            "groups",
            nargs='+',
            metavar="group",
            help="Group(s) dropping the role",
        )

    @override
    def cleanup_arguments(self, *args: Any, **options: Any) -> None:
        if "data_file" in options and options["data_file"] != sys.stdin:
            options["data_file"].close()

    def get_asset_by_id(self, id_: int | str) -> Asset:
        """Lookup an asset by ID."""
        try:
            return Asset.objects.get(id=int(id_))
        except Asset.DoesNotExist:
            raise CommandError(f"Asset ID {id_!r} not found", returncode=3)

    def get_group(self, scope: Scope, group_name: str) -> Group:
        """Lookup a group in a scope."""
        try:
            group = Group.objects.get(scope=scope, name=group_name)
        except Group.DoesNotExist:
            raise CommandError(
                f"Group {group_name!r} not found in scope {scope.name!r}",
                returncode=3,
            )
        return group

    def get_role_by_name(self, role_name: str, usage_role: bool) -> str:
        """Return the Roles object referring to role_name."""
        role_class = AssetUsageRoles if usage_role else AssetRoles
        try:
            return role_class(role_name)
        except ValueError as e:
            raise CommandError(e, returncode=3)

    def assert_workspace_in_scope(
        self, workspace: Workspace, scope: Scope
    ) -> None:
        """Ensure that workspace is within scope."""
        if scope != workspace.scope:
            raise CommandError(
                f"{workspace!r} is not in {scope!r}",
                returncode=3,
            )

    @override
    @context.disable_permission_checks()
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Dispatch the requested action."""
        func = cast(
            Callable[..., NoReturn],
            getattr(self, f"handle_{options['action']}", None),
        )
        func(*args, **options)

    def handle_create(
        self,
        category: str,
        workspace: Workspace | None,
        data_file: IO[Any],
        **options: Any,
    ) -> NoReturn:
        """Create or update an asset."""
        data = self.parse_yaml_data(data_file.read()) or {}

        # For idempotency, we try to match the known unique constraints for
        # various asset categories.
        filters: dict[str, Any] = {}
        match category:
            case AssetCategory.CLOUD_PROVIDER_ACCOUNT:
                filters["data__name"] = data.get("name")
            case AssetCategory.SIGNING_KEY:
                filters["data__fingerprint"] = data.get("fingerprint")
            case _:  # pragma: no cover
                # No other asset categories are defined yet.
                pass

        with transaction.atomic():
            try:
                try:
                    asset = Asset.objects.get(
                        category=category, workspace=workspace, **filters
                    )
                except Asset.DoesNotExist:
                    asset = Asset.objects.create(
                        category=category, workspace=workspace, data=data
                    )
                if asset.data != data:
                    asset.data = data
                # TODO: We can't validate constraints here, because
                # JsonDataUniqueConstraint doesn't implement validate().  If
                # there's a constraint violation then we'll get an
                # IntegrityError instead.
                asset.full_clean(validate_constraints=False)
                asset.save()
            except ValidationError as exc:
                raise CommandError(
                    "Error creating asset: " + "\n".join(exc.messages),
                    returncode=3,
                )
            except IntegrityError as exc:
                raise CommandError(f"Error creating asset: {exc}", returncode=3)

        raise SystemExit(0)

    def handle_list(
        self,
        scope: Scope | None,
        workspace: Workspace | None,
        yaml: bool,
        **options: Any,
    ) -> NoReturn:
        """List assets in a scope."""
        if scope and workspace:
            self.assert_workspace_in_scope(workspace, scope)
        assets = Asset.objects.all()
        if scope is not None:
            assets = assets.filter(workspace__scope=scope)
        if workspace is not None:
            assets = assets.filter(workspace=workspace)
        Assets(yaml).print(assets, self.stdout)
        raise SystemExit(0)

    def handle_list_roles(
        self, asset: Asset, yaml: bool, **options: Any
    ) -> NoReturn:
        """List assets's roles."""
        ResourceRoles(yaml).print(asset.roles.all(), self.stdout)
        usage_roles = AssetUsageRole.objects.filter(resource__asset=asset).all()
        if yaml:
            print("---")
        AssetUsageRolesPrinter(yaml).print(usage_roles, self.stdout)
        raise SystemExit(0)

    def handle_grant_role(
        self,
        asset: Asset,
        role_name: str,
        groups: list[str],
        workspace: Workspace | None,
        **options: Any,
    ) -> NoReturn:
        """Assign role on asset to groups."""
        role = self.get_role_by_name(
            role_name, usage_role=workspace is not None
        )
        resource: Asset | AssetUsage = asset
        if workspace:
            if asset.workspace:
                self.assert_workspace_in_scope(workspace, asset.workspace.scope)
            resource, _ = AssetUsage.objects.get_or_create(
                asset=asset, workspace=workspace
            )
        if asset.workspace:
            scope = asset.workspace.scope
        elif workspace:
            scope = workspace.scope
        else:
            raise CommandError(
                "Cannot grant roles on an asset not assigned to a workspace"
            )
        for group_name in groups:
            group = self.get_group(scope, group_name)
            group.assign_role(resource, role)
        raise SystemExit(0)

    def handle_revoke_role(
        self,
        asset: Asset,
        workspace: Workspace | None,
        role_name: str,
        groups: list[str],
        **options: Any,
    ) -> NoReturn:
        """Revoke role on asset from groups."""
        if asset.workspace:
            scope = asset.workspace.scope
        elif workspace:
            scope = workspace.scope
        else:
            raise CommandError(
                "Cannot revoke roles on an asset not assigned to a workspace"
            )
        if workspace and asset.workspace:
            self.assert_workspace_in_scope(workspace, asset.workspace.scope)
        role = self.get_role_by_name(
            role_name, usage_role=workspace is not None
        )
        for group_name in groups:
            group = self.get_group(scope, group_name)
            if workspace:
                AssetUsageRole.objects.filter(
                    resource__asset=asset, group=group, role=role
                ).delete()
            else:
                AssetRole.objects.filter(
                    resource=asset, group=group, role=role
                ).delete()
        raise SystemExit(0)

    def handle_delete(
        self, *, asset: Asset, yes: bool, **options: Any
    ) -> NoReturn:
        """Delete an asset with associated resources."""
        deletion_confirmed = False
        if yes:
            deletion_confirmed = True
        else:
            deletion_answer = input(
                f"Would you like to delete asset {asset}? [yN] "
            )
            deletion_confirmed = deletion_answer.strip() in ('y', 'Y')

        if not deletion_confirmed:
            self.stderr.write("Not deleted.\n")
            raise SystemExit(1)

        asset.delete()
        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the create_child_workspace workflow."""

from debusine.db.models import WorkRequest, Workspace
from debusine.db.playground import scenarios
from debusine.server.workflows import CreateChildWorkspaceWorkflow
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import ChildWorkspaceData
from debusine.server.workflows.tests.helpers import WorkflowTestCase
from debusine.tasks.models import BaseDynamicTaskData


class CreateChildWorkspaceWorkflowTests(
    WorkflowTestCase[CreateChildWorkspaceWorkflow]
):
    """Unit tests for :py:class:`CreateChildWorkspaceWorkflow`."""

    SAMPLE_TASK_DATA = {"suffix": "test"}

    scenario = scenarios.DefaultContext()

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BaseDynamicTaskData(subject="test", parameter_summary="test"),
        )

    def test_get_label_without_parameter_summary(self) -> None:
        """Test get_label."""
        self.task_data["suffix"] = "test_1"
        self.assertEqual(self.task.get_label(), "create_child_workspace")

    def test_get_label_with_parameter_summary(self) -> None:
        """Test get_label."""
        root = self.orchestrate(task_data=ChildWorkspaceData(suffix="test_2"))

        self.assertEqual(
            root.get_task().get_label(),
            "create_child_workspace-template: test_2",
        )

    def test_create_orchestrator(self) -> None:
        """Workflow can be instantiated."""
        workflow = self.task
        self.assertIsNone(workflow.data.prefix)
        self.assertEqual(workflow.data.suffix, "test")
        self.assertTrue(workflow.data.public)
        self.assertIsNone(workflow.data.owner_group)
        self.assertEqual(workflow.data.workflow_template_names, [])
        self.assertEqual(workflow.data.expiration_delay, 60)

    def orchestrate(
        self,
        task_data: ChildWorkspaceData,
        *,
        workspace: Workspace | None = None,
    ) -> WorkRequest:
        """Create and orchestrate a CreateChildWorkspaceWorkflow."""
        template = self.playground.create_workflow_template(
            name="create_child_workspace-template",
            task_name="create_child_workspace",
            workspace=workspace,
        )
        root = self.playground.create_workflow(
            task_name="create_child_workspace",
            task_data=task_data.model_dump(mode="json"),
        )
        self.assertEqual(root.workspace, template.workspace)
        self.assertTrue(orchestrate_workflow(root))
        return root

    def test_populate(self) -> None:
        root = self.orchestrate(task_data=ChildWorkspaceData(suffix="test"))
        # There is only one child task
        create_child_workspace = root.children.get()
        self.assertEqual(
            create_child_workspace.task_name, "create_child_workspace"
        )
        self.assertEqual(
            create_child_workspace.task_data,
            {
                "prefix": root.workspace.name,
                "suffix": "test",
                "expiration_delay": 60,
                "owner_group": None,
                "public": True,
                "workflow_template_names": [],
            },
        )

    def test_populate_prefix(self) -> None:
        root = self.orchestrate(
            task_data=ChildWorkspaceData(prefix="namespace", suffix="test")
        )
        # There is only one child task
        create_child_workspace = root.children.get()
        self.assertEqual(
            create_child_workspace.task_name, "create_child_workspace"
        )
        self.assertEqual(
            create_child_workspace.task_data,
            {
                "prefix": "namespace",
                "suffix": "test",
                "expiration_delay": 60,
                "owner_group": None,
                "public": True,
                "workflow_template_names": [],
            },
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Tests for rest framework extensions."""

from typing import override

from django.http import HttpRequest
from rest_framework.request import Request
from rest_framework.views import APIView

from debusine.server.views.rest import (
    IsGet,
    IsTokenAuthenticated,
    IsUserTokenAuthenticated,
    IsWorkRequestTokenAuthenticated,
    IsWorkerTokenAuthenticated,
)
from debusine.test.django import TestCase


class IsGetTests(TestCase):
    """Tests for IsGet."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.is_get = IsGet()
        self.http_request = Request(HttpRequest())

    def test_request_is_get(self) -> None:
        """Request method is GET. Return True."""
        self.http_request.method = "GET"

        self.assertTrue(
            self.is_get.has_permission(self.http_request, APIView())
        )

    def test_request_is_post(self) -> None:
        """Request method is POST. Return False."""
        self.http_request.method = "POST"

        self.assertFalse(
            self.is_get.has_permission(self.http_request, APIView())
        )


class IsTokenAuthenticatedTests(TestCase):
    """Tests for IsTokenAuthenticated."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.is_token_authenticated = IsTokenAuthenticated()
        self.request = Request(HttpRequest())

    def test_request_with_valid_token_worker_yes_permission(self) -> None:
        """IsTokenAuthenticated.has_permission() return True: valid token."""
        self.request.auth = self.playground.create_worker_token()
        self.assertTrue(
            self.is_token_authenticated.has_permission(self.request, APIView())
        )

    def test_request_with_non_existing_token_no_permission(self) -> None:
        """IsTokenAuthenticated.has_permission() return False: invalid token."""
        self.request.auth = None
        self.assertFalse(
            self.is_token_authenticated.has_permission(self.request, APIView())
        )

    def test_request_with_disabled_token_no_permission(self) -> None:
        """
        IsTokenAuthenticated.has_permission() return False.

        The token is disabled.
        """
        self.request.auth = self.playground.create_worker_token(enabled=False)
        self.assertFalse(
            self.is_token_authenticated.has_permission(self.request, APIView())
        )


class IsUserTokenAuthenticatedTests(TestCase):
    """Tests for IsUserTokenAuthenticated class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.permission = IsUserTokenAuthenticated()
        self.request = Request(HttpRequest())

    def test_request_disabled_token(self) -> None:
        """Request with a Token header that is disabled: permission denied."""
        self.request.auth = self.playground.create_user_token(enabled=False)
        self.assertFalse(self.request.auth.enabled)
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_request_non_existing_token(self) -> None:
        """Request without a token."""
        self.request.auth = None
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_request_worker_token(self) -> None:
        """Request with a Worker Token: permission denied."""
        self.request.auth = self.playground.create_worker_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_request_worker_activation_token(self) -> None:
        """Request with a Worker Activation Token: permission denied."""
        self.request.auth = self.playground.create_worker_activation_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_request_enabled_token_with_user(self) -> None:
        """Request with a Token that does have a user: permission granted.."""
        self.request.auth = self.playground.create_user_token()
        self.assertTrue(self.permission.has_permission(self.request, APIView()))

    def test_request_work_request_token(self) -> None:
        """Request with a Work Request scoped Token: permission denied."""
        self.request.auth = self.playground.create_work_request_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )


class IsWorkerTokenAuthenticatedTests(TestCase):
    """Tests for IsWorkerTokenAuthenticated."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.permission = IsWorkerTokenAuthenticated()
        self.request = Request(HttpRequest())

    def test_request_with_valid_token_worker_yes_permission(self) -> None:
        """IsWorkerTokenAuthenticated accepts valid token."""
        self.request.auth = self.playground.create_worker_token()
        self.assertTrue(self.permission.has_permission(self.request, APIView()))

    def test_request_with_non_existing_token_no_permission(self) -> None:
        """
        IsWorkerTokenAuthenticated.has_permission() return False.

        request.auth is None.
        """
        self.request.auth = None
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_request_with_user_token_no_permission(self) -> None:
        """IsWorkerTokenAuthenticated rejects a user token."""
        self.request.auth = self.playground.create_user_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_request_work_request_token(self) -> None:
        """IsWorkerTokenAuthenticated rejects a work request scoped token."""
        self.request.auth = self.playground.create_work_request_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_request_worker_activation_token(self) -> None:
        """IsWorkerTokenAuthenticated rejects a worker activation token."""
        self.request.auth = self.playground.create_worker_activation_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )


class IsWorkRequestTokenAuthenticatedTests(TestCase):
    """Tests for IsWorkRequestTokenAuthenticated."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.permission = IsWorkRequestTokenAuthenticated()
        self.request = Request(HttpRequest())

    def test_has_permission_valid_token(self) -> None:
        """has_permission() accepts a valid token."""
        self.request.auth = self.playground.create_work_request_token()
        self.assertTrue(self.permission.has_permission(self.request, APIView()))

    def test_has_permission_no_token(self) -> None:
        """has_permission() rejects an unauthenticated request."""
        self.request.auth = None
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_has_permission_worker_token(self) -> None:
        """has_permission() rejects a non-scoped worker token."""
        self.request.auth = self.playground.create_worker_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_has_permission_user_token(self) -> None:
        """has_permission() rejects a user token."""
        self.request.auth = self.playground.create_user_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )

    def test_has_permission_worker_activation_token(self) -> None:
        """has_permission() rejects a worker activation token."""
        self.request.auth = self.playground.create_worker_activation_token()
        self.assertFalse(
            self.permission.has_permission(self.request, APIView())
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command create_notification_channel."""

import io
from typing import ClassVar, override

import yaml
from django.core.management import CommandError

from debusine.db.models import NotificationChannel
from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


class CreateNotificationChannelCommandTests(TestCase):
    """Tests for the create_notification_channel command."""

    yaml_data: ClassVar[str]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.yaml_data = yaml.safe_dump(
            {"from": "debusine@debusine.com", "to": ["lts@debusine.com"]}
        )

    def test_create_notification_channel_from_file(self) -> None:
        """create_notification_channel create a notification (data in file)."""
        name = "lts"
        method = NotificationChannel.Methods.EMAIL
        data_file = self.create_temporary_file(
            contents=self.yaml_data.encode("utf-8")
        )
        stdout, stderr, exit_code = call_command(
            "create_notification_channel",
            name,
            method,
            "--data",
            str(data_file),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertTrue(
            NotificationChannel.objects.filter(
                name=name, method=method
            ).exists()
        )

    def test_create_notification_channel_from_stdin(self) -> None:
        """create_notification_channel create a notification (data in stdin)."""
        name = "lts"
        method = NotificationChannel.Methods.EMAIL

        stdout, stderr, exit_code = call_command(
            "create_notification_channel",
            name,
            method,
            stdin=io.StringIO(self.yaml_data),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertTrue(
            NotificationChannel.objects.filter(
                name=name, method=method
            ).exists()
        )

    def test_create_notification_channel_invalid_data_yaml(self) -> None:
        """create_notification_channel return error: cannot parse YAML data."""
        data_file = self.create_temporary_file(contents=b":")
        with self.assertRaisesRegex(
            CommandError, "^Error parsing YAML:"
        ) as exc:
            call_command(
                "create_notification_channel",
                "test",
                "email",
                "--data",
                str(data_file),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_create_notification_invalid_data(self) -> None:
        """create_notification_channel return error: data is invalid."""
        # it is invalid because does not have the required fields
        data_file = self.create_temporary_file(contents=b"")
        with self.assertRaisesRegex(CommandError, "^Error creating") as exc:
            call_command(
                "create_notification_channel",
                "test",
                "email",
                "--data",
                str(data_file),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_create_notification_invalid_name(self) -> None:
        """create_notification_channel return error: name is invalid."""
        # it is invalid because does not have the required fields
        data_file = self.create_temporary_file(contents=b"{}")
        with self.assertRaisesRegex(CommandError, "^Error creating") as exc:
            call_command(
                "create_notification_channel",
                "a: b",
                "email",
                "--data",
                str(data_file),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_create_duplicated_notification_raise_command_error(self) -> None:
        """create_notification_channel raise CommandError: duplicated name."""
        name = "lts"
        method = NotificationChannel.Methods.EMAIL
        data_file = self.create_temporary_file(
            contents=self.yaml_data.encode("utf-8")
        )

        call_command(
            "create_notification_channel",
            name,
            method,
            "--data",
            str(data_file),
        )

        with self.assertRaisesRegex(CommandError, "^Error creating") as exc:
            call_command(
                "create_notification_channel",
                name,
                method,
                "--data",
                str(data_file),
            )

        self.assertEqual(exc.exception.returncode, 3)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the scope views."""

import datetime as dt
from typing import ClassVar, override

import lxml
from django.urls import reverse
from rest_framework import status

from debusine.db.models import Scope, Workspace
from debusine.db.playground import scenarios
from debusine.server.scopes import urlconf_scope
from debusine.test.django import TestCase
from debusine.web.views.tests.utils import ViewTestMixin


class ScopeDetailViewTestsEmptyDB(ViewTestMixin, TestCase):
    """Tests for ScopeDetailView with no workspaces."""

    def test_no_workspaces(self) -> None:
        """No workspaces: 'No workspaces' in response."""
        default_workspace = self.playground.get_default_workspace()
        default_workspace.public = False
        default_workspace.save()

        response = self.client.get(reverse("scopes:detail"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertContains(response, "No workspaces.", html=True)


class ScopeDetailViewTests(ViewTestMixin, TestCase):
    """Tests for the ScopeDetailView class."""

    scenario = scenarios.DefaultContext()
    scope: ClassVar[Scope]
    private_workspace: ClassVar[Workspace]
    scope_workspace: ClassVar[Workspace]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope = cls.playground.get_or_create_scope("scope")
        cls.private_workspace = cls.playground.create_workspace(name="test")
        cls.scope_workspace = cls.playground.create_workspace(
            scope=cls.scope, public=True, name="other"
        )

    def assertWorkspaces(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        workspaces: list[Workspace],
    ) -> None:
        """Check that a tr element is a workspace list table row."""
        rows = self.workspace_list_table_rows(tree)
        for tr, workspace in zip(rows, workspaces):
            self.assertEqual(tr.tag, "tr")
            a = tr.td[0].a
            self.assertEqual(a.text, workspace.name)
        self.assertEqual(len(rows), len(workspaces))

    def test_reverse(self) -> None:
        """Test reverse interacting with scopes."""
        # Without context set, default to DEBUSINE_DEFAULT_SCOPE
        self.assertEqual(reverse("scopes:detail"), "/debusine/")

        self.scenario.set_current()
        self.assertEqual(reverse("scopes:detail"), "/debusine/")

        # Simulate a different request.urlconf
        with urlconf_scope("scope"):
            self.assertEqual(reverse("scopes:detail"), "/scope/")

    def test_view(self) -> None:
        """Scope view loads."""
        self.scenario.set_current()
        response = self.client.get(reverse("scopes:detail"))
        tree = self.assertResponseHTML(response)

        title = self.assertHasElement(tree, "//title")
        self.assertTextContentEqual(title, f"{self.scenario.scope.name}")

        h1 = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(h1, self.scenario.scope.name)

        # self.private_workspace is not shown
        self.assertWorkspaces(tree, [self.scenario.workspace])

    def test_view_anonymous(self) -> None:
        """Ensure anonymous users only see public workspaces."""
        response = self.client.get(reverse("scopes:detail"))
        tree = self.assertResponseHTML(response)
        # Only public workspaces are shown
        self.assertWorkspaces(tree, [self.scenario.workspace])

    def test_view_with_private_access(self) -> None:
        """Ensure contributors can see their workspaces."""
        self.playground.create_group_role(
            self.private_workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.scenario.user],
        )
        self.client.force_login(self.scenario.user)
        response = self.client.get(reverse("scopes:detail"))
        tree = self.assertResponseHTML(response)
        self.assertWorkspaces(
            tree, [self.scenario.workspace, self.private_workspace]
        )

    def test_view_otherscope(self) -> None:
        """Test listing workspaces on a different scope."""
        with urlconf_scope("scope"):
            url = reverse("scopes:detail")
        response = self.client.get(url)
        tree = self.assertResponseHTML(response)

        title = self.assertHasElement(tree, "//title")
        self.assertTextContentEqual(title, f"{self.scope.name}")

        h1 = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(h1, self.scope.name)

        self.assertWorkspaces(tree, [self.scope_workspace])

    def test_list_workspace_expiration(self) -> None:
        """Render public workspace's non-NULL default expiration delay."""
        self.scenario.workspace.default_expiration_delay = dt.timedelta(400)
        self.scenario.workspace.save()
        response = self.client.get(reverse("scopes:detail"))
        tree = self.assertResponseHTML(response)
        rows = self.workspace_list_table_rows(tree)
        self.assertTextContentEqual(rows[0].td[2], "400")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Exceptions used by debusine.client."""

from collections.abc import Iterable
from typing import Any, override


class NotFoundError(Exception):
    """
    Raised if the server returns 404.

    The reason is probably that the queried item cannot be found. It could
    also be that the client requested a non-existing URL. Check message for
    more information.
    """


class UnexpectedResponseError(Exception):
    """
    Raised if the client received an unexpected response from the server.

    It could be a non-expected status code, the response is not JSON, etc.
    """


class ClientForbiddenError(Exception):
    """Raise if the server return 403 to a client."""


class ClientConnectionError(Exception):
    """Raised when the server cannot be reached: timeout, unreachable, etc."""


class TokenDisabledError(Exception):
    """Raised if the server disconnects the worker: the token is disabled."""


class DebusineError(Exception):
    """
    Raised when the debusine server returns HTTP 400 with a message.

    The message is in the JSON content in "detail".
    """

    def __init__(
        self,
        title: str,
        detail: str | None = None,
        validation_errors: Iterable[Any] | None = None,
        status_code: int | None = None,
    ) -> None:
        """Initialize the exception."""
        self.title = title
        self.detail = detail
        self.validation_errors = validation_errors
        self.status_code = status_code

    @override
    def __str__(self) -> str:
        parts: list[str] = []
        if self.status_code is not None:
            parts.append(f"{self.status_code} ")
        parts.append(self.title)
        if self.detail is not None:
            parts.append(f": {self.detail}")
        return "".join(parts)

    def asdict(self) -> dict[str, Any]:
        """Return the DebusineError as dictionary."""
        res: dict[str, Any] = {
            "title": self.title,
        }
        if self.detail is not None:
            res["detail"] = self.detail
        if self.validation_errors is not None:
            res["validation_errors"] = self.validation_errors
        if self.status_code is not None:
            res["status_code"] = self.status_code
        return res


class ContentValidationError(Exception):
    """Raised if downloaded data doesn't match the expected hash or size."""


class ContentTooLargeError(Exception):
    """Raised if content is bigger than expected."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Signing service data models."""

import base64
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable
from contextlib import contextmanager
from enum import StrEnum
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Annotated, Literal, Self, TYPE_CHECKING, override

import pydantic
from django.conf import settings

from debusine.signing.signing_utils import SensitiveTemporaryDirectory
from debusine.utils import Base64BytesAsStr

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer
    from nacl.public import PrivateKey, PublicKey

    ReadableBuffer  # fake usage for vulture


class ProtectedKeyError(Exception):
    """An error handling a protected key."""


class ProtectedKeyStorage(StrEnum):
    """Possible values for ProtectedKey.storage."""

    NACL = "nacl"
    PKCS11_STATIC = "pkcs11-static"


class BaseAvailableKey(ABC):
    """A protected key that has temporarily been made available."""

    def __init__(
        self, *, temp_dir: Path, certificate: "ReadableBuffer"
    ) -> None:
        """Construct a :py:class:`BaseAvailableKey`."""
        self.temp_dir = temp_dir
        self._certificate_path.write_bytes(certificate)

    @property
    def _certificate_path(self) -> Path:
        return self.temp_dir / "tmp.crt"

    @property
    @abstractmethod
    def sbsign_args(self) -> list[str]:
        """Arguments for `sbsign`."""


class BaseProtectedKey(pydantic.BaseModel, ABC):
    """Base class for protected private keys."""

    storage: ProtectedKeyStorage

    @contextmanager
    @abstractmethod
    def available(
        self, public_key: "ReadableBuffer"
    ) -> Generator[BaseAvailableKey]:
        """
        Make a protected key temporarily available.

        Controls on the available key may vary depending on the type of
        protection used, but they should at least ensure that private keys
        do not end up on a persistent file system.
        """


class AvailableKeyFileSystem(BaseAvailableKey):
    """An available private key on the file system."""

    def __init__(
        self,
        *,
        temp_dir: Path,
        certificate: "ReadableBuffer",
        key: "ReadableBuffer",
    ) -> None:
        """Construct an :py:class:`AvailableKeyFileSystem`."""
        super().__init__(temp_dir=temp_dir, certificate=certificate)
        self._key_path.write_bytes(key)

    @property
    def _key_path(self) -> Path:
        return self.temp_dir / "tmp.key"

    @override
    @property
    def sbsign_args(self) -> list[str]:
        return [
            "--key",
            str(self._key_path),
            "--cert",
            str(self._certificate_path),
        ]


class AvailableKeyPKCS11(BaseAvailableKey):
    """An available private key on a PKCS#11 token."""

    def __init__(
        self, *, temp_dir: Path, certificate: "ReadableBuffer", pkcs11_uri: str
    ) -> None:
        """Construct an :py:class:`AvailableKeyPKCS11`."""
        super().__init__(temp_dir=temp_dir, certificate=certificate)
        self.pkcs11_uri = pkcs11_uri

    @override
    @property
    def sbsign_args(self) -> list[str]:
        return [
            "--engine",
            "pkcs11",
            "--key",
            self.pkcs11_uri,
            "--cert",
            str(self._certificate_path),
        ]


class ProtectedKeyNaCl(BaseProtectedKey):
    """Data for a private key encrypted in software using NaCl."""

    storage: Literal[ProtectedKeyStorage.NACL]
    public_key: Base64BytesAsStr
    encrypted: Base64BytesAsStr

    @classmethod
    def create(
        cls, *, public_key: "ReadableBuffer", encrypted: "ReadableBuffer"
    ) -> Self:
        """Create a new :py:class:`ProtectedKeyNaCl`."""
        return cls(
            storage=ProtectedKeyStorage.NACL,
            public_key=base64.b64encode(public_key).decode(),
            encrypted=base64.b64encode(encrypted).decode(),
        )

    @classmethod
    def encrypt(cls, public_key: "PublicKey", data: bytes) -> Self:
        """Encrypt data using a NaCl public key."""
        from nacl.public import SealedBox

        try:
            encrypted = SealedBox(public_key).encrypt(data)
        except Exception as e:
            raise ProtectedKeyError(str(e)) from e
        return cls.create(public_key=bytes(public_key), encrypted=encrypted)

    def decrypt(self, private_keys: Iterable["PrivateKey"]) -> bytes:
        """
        Decrypt data using any of an iterable of NaCl private keys.

        This uses the private key that matches the stored public key, if one
        exists.  This allows for key rotation.

        :raises ProtectedKeyError: if none of the given private keys match
          the stored public key.
        """
        from nacl.exceptions import CryptoError
        from nacl.public import SealedBox

        for private_key in private_keys:
            if bytes(private_key.public_key) == self.public_key:
                try:
                    return SealedBox(private_key).decrypt(self.encrypted)
                except CryptoError as e:
                    raise ProtectedKeyError(str(e)) from e
        raise ProtectedKeyError(
            "Key not encrypted using any of the given private keys"
        )

    @override
    @contextmanager
    def available(
        self, public_key: "ReadableBuffer"
    ) -> Generator[BaseAvailableKey]:
        """Make this protected key temporarily available."""
        with SensitiveTemporaryDirectory(
            "debusine-available-nacl-"
        ) as temp_dir:
            yield AvailableKeyFileSystem(
                temp_dir=Path(temp_dir),
                certificate=public_key,
                key=self.decrypt(settings.DEBUSINE_SIGNING_PRIVATE_KEYS),
            )


class ProtectedKeyPKCS11Static(BaseProtectedKey):
    """
    Data for a private key held on a PKCS#11 token.

    This key is not extracted under wrap; it only ever lives on the token.
    """

    storage: Literal[ProtectedKeyStorage.PKCS11_STATIC]
    pkcs11_uri: str

    @classmethod
    def create(cls, *, pkcs11_uri: str) -> Self:
        """Create a new :py:class:`ProtectedKeyPKCS11Static`."""
        return cls(
            storage=ProtectedKeyStorage.PKCS11_STATIC, pkcs11_uri=pkcs11_uri
        )

    @pydantic.field_validator("pkcs11_uri", mode="after")
    @classmethod
    def _pkcs11_uri_is_pkcs11(cls, value: str) -> str:
        """Ensure that the PKCS#11 URI uses the "pkcs11" scheme."""
        if not value.startswith("pkcs11:"):
            raise ValueError("pkcs11_uri must start with 'pkcs11:'")
        return value

    @override
    @contextmanager
    def available(
        self, public_key: "ReadableBuffer"
    ) -> Generator[BaseAvailableKey]:
        """Make this protected key temporarily available."""
        with TemporaryDirectory(
            "debusine-available-pkcs11-static-"
        ) as temp_dir:
            yield AvailableKeyPKCS11(
                temp_dir=Path(temp_dir),
                certificate=public_key,
                pkcs11_uri=self.pkcs11_uri,
            )


class ProtectedKey(
    pydantic.RootModel[
        Annotated[
            ProtectedKeyNaCl | ProtectedKeyPKCS11Static,
            pydantic.Field(discriminator="storage"),
        ]
    ]
):
    """A protected private key."""

    @contextmanager
    def available(
        self, public_key: "ReadableBuffer"
    ) -> Generator[BaseAvailableKey]:
        """
        Make a protected key temporarily available.

        Controls on the available key may vary depending on the type of
        protection used, but they should at least ensure that private keys
        do not end up on a persistent file system.
        """
        with self.root.available(public_key) as available:
            yield available


class SigningMode(StrEnum):
    """
    The kind of signing to perform.

    Only some combinations of key purpose and signing mode are valid.
    """

    # Return the original data with an attached signature.
    ATTACHED = "attached"

    # Return a detached signature.
    DETACHED = "detached"

    # Return the original data with an attached signature, encapsulated as
    # plain text.  This is only valid for OpenPGP.
    CLEAR = "clear"

    # Sign a complete Debian upload using `debsign` or equivalent.
    DEBSIGN = "debsign"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Wait task to delay until a given time is reached."""

from typing import override

from debusine.server.tasks.wait import BaseWaitTask
from debusine.server.tasks.wait.models import DelayData


class Delay(BaseWaitTask[DelayData]):
    """Task that delays until a given time is reached."""

    TASK_VERSION = 1

    @override
    def get_label(self) -> str:
        return f"delay until {self.data.delay_until}"



# Generated by Django 4.2.23 on 2025-12-23 17:08

from django.db import migrations, models

import debusine.tasks.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0038_worker_tombstones_without_tokens'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='worker',
            name='db_worker_celery_or_token',
        ),
        migrations.AlterField(
            model_name='worker',
            name='worker_type',
            field=models.CharField(
                choices=[
                    (
                        debusine.tasks.models.WorkerType['SCHEDULER'],
                        debusine.tasks.models.WorkerType['SCHEDULER'],
                    ),
                    (
                        debusine.tasks.models.WorkerType['EXTERNAL'],
                        debusine.tasks.models.WorkerType['EXTERNAL'],
                    ),
                    (
                        debusine.tasks.models.WorkerType['CELERY'],
                        debusine.tasks.models.WorkerType['CELERY'],
                    ),
                    (
                        debusine.tasks.models.WorkerType['SIGNING'],
                        debusine.tasks.models.WorkerType['SIGNING'],
                    ),
                ],
                default=debusine.tasks.models.WorkerType['EXTERNAL'],
                editable=False,
            ),
        ),
        migrations.AddConstraint(
            model_name='worker',
            constraint=models.CheckConstraint(
                check=models.Q(
                    (
                        'worker_type__in',
                        {
                            debusine.tasks.models.WorkerType['SCHEDULER'],
                            debusine.tasks.models.WorkerType['CELERY'],
                        },
                    ),
                    ('activation_token__isnull', False),
                    ('token__isnull', False),
                    models.Q(
                        ('instance_created_at__isnull', True),
                        ('worker_pool_id__isnull', False),
                    ),
                    _connector='OR',
                ),
                name='db_worker_celery_or_token',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Local artifact's representation.

See docs/design/ontology.html for the semantics of the Artifacts.
"""

import abc
import json
import re
from collections.abc import Iterable, Sequence
from json import JSONDecodeError
from pathlib import Path, PurePath
from typing import Any, ClassVar, Self, assert_never, overload, override

import pydantic
from debian import deb822, debfile

# https://github.com/pydantic/pydantic/issues/12720
from pydantic._internal._generics import get_origin

import debusine.artifacts.models as data_models
from debusine import utils
from debusine.artifacts.artifact_utils import files_in_meta_file_match_files
from debusine.assets import KeyPurpose
from debusine.client.models import FileRequest
from debusine.utils import (
    extract_generic_type_arguments,
    is_single_path_component,
    read_changes,
)


class NewLocalFile(pydantic.BaseModel):
    """A new local file to be added to the artifact."""

    #: File in the local file system to be added to the artifact.
    file: Path
    #: Base directory of the artifact.  Must be an absolute path.  If it's
    #: None, the file is added in the root of the artifact; otherwise, it is
    #: added with the path of the file relative to `artifact_base_dir`.
    artifact_base_dir: Path | None = None
    #: Name to use for the file in the artifact instead of `file.name`.
    override_name: str | None = None
    #: Content-Type to use when serving the file instead of guessing.
    content_type: str | None = None


class NewRemoteFile(pydantic.BaseModel):
    """
    A new remote file to be added to the artifact.

    This may be used when a file does not exist locally, but can be expected
    to already exist on the server when uploading the rest of the artifact.
    """

    #: Path of the file in the artifact.
    path_in_artifact: str
    #: Description of the remote file.
    file_request: FileRequest


class NascentLocalArtifact(pydantic.BaseModel):
    """Information about files that will be added to a new local artifact."""

    local_files: dict[str, Path] = pydantic.Field(default_factory=dict)
    remote_files: dict[str, FileRequest] = pydantic.Field(default_factory=dict)
    content_types: dict[str, str] = pydantic.Field(default_factory=dict)

    def _check_existing_file(
        self, path_in_artifact: str, new_file: Path | FileRequest
    ) -> None:
        """Check whether the nascent artifact already contains a given file."""
        all_files = self.local_files | self.remote_files
        if path_in_artifact in all_files:
            raise ValueError(
                f"File with the same path ({path_in_artifact}) "
                f"is already in the artifact "
                f'("{all_files[path_in_artifact]}" and "{new_file}")'
            )

    def add_local_file(self, new_file: NewLocalFile) -> None:
        """
        Prepare a local file for addition to the artifact.

        :param new_file: information about the new local file.
        :raises ValueError: artifact_base_dir is not absolute or is not
          a directory; file does not exist; the path in the artifact already
          had a file.
        """
        file = new_file.file
        if new_file.artifact_base_dir is not None:
            if not new_file.artifact_base_dir.is_absolute():
                raise ValueError(
                    f'"{new_file.artifact_base_dir}" must be absolute'
                )
            if not new_file.artifact_base_dir.is_dir():
                raise ValueError(
                    f'"{new_file.artifact_base_dir}" does not exist or '
                    f'is not a directory'
                )

            if not file.is_absolute():
                file = new_file.artifact_base_dir.joinpath(file)

            path_in_artifact = file.relative_to(
                new_file.artifact_base_dir
            ).as_posix()
        else:
            path_in_artifact = file.name

        if new_file.override_name is not None:
            path_in_artifact = new_file.override_name

        if not file.exists():
            raise ValueError(f'"{file}" does not exist')

        if not file.is_file():
            raise ValueError(f'"{file}" is not a file')

        file_absolute = file.absolute()

        self._check_existing_file(path_in_artifact, file_absolute)
        self.local_files[path_in_artifact] = file_absolute
        if new_file.content_type is not None:
            self.content_types[path_in_artifact] = new_file.content_type

    def add_remote_file(self, new_file: NewRemoteFile) -> None:
        """
        Prepare a remote file for addition to the artifact.

        :param new_file: information about the new remote file.
        """
        self._check_existing_file(
            new_file.path_in_artifact, new_file.file_request
        )
        self.remote_files[new_file.path_in_artifact] = new_file.file_request


class LocalArtifact[AD: data_models.ArtifactData](pydantic.BaseModel, abc.ABC):
    """Represent an artifact locally."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")

    #: Artifact type
    category: data_models.ArtifactCategory
    # Keys are paths in the artifact. Values the paths in the local system
    files: dict[str, Path] = pydantic.Field(default_factory=dict)
    # Keys are paths in the artifact. Values are FileRequests for files that do
    # not exist locally but can be expected to already exist on the server.
    remote_files: dict[str, FileRequest] = pydantic.Field(default_factory=dict)
    # Keys are paths in the artifact. Values are MIME types, suitable for
    # Content-Type headers.
    content_types: dict[str, str] = pydantic.Field(default_factory=dict)
    #: Artifact data
    data: AD
    # TODO: it would be great to not have to redefine data in subclasses, but
    # it needs pydantic's Generics support, which might require work to work
    # with extract_generic_type_arguments

    #: Default value for category
    _category: ClassVar[data_models.ArtifactCategory]

    #: Class used as the in-memory representation of artifact data.
    _data_type: type[AD] = pydantic.PrivateAttr()
    # data_type is marked as PrivateAttr to make mypy happy. Setting
    # underscore_attrs_are_private to True does not seem to be enough

    _local_artifacts_category_to_class: ClassVar[
        dict[str, type["LocalArtifact['Any']"]]
    ] = {}

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Register subclass into LocalArtifact._local_artifacts_category_to_class.

        Allow to list possible valid options (in the client or server).
        """
        super().__init_subclass__(**kwargs)

        if get_origin(cls) is not None:
            # The task data type, computed by introspecting the type argument
            # used to specialize this generic class.
            [cls._data_type] = extract_generic_type_arguments(
                cls, LocalArtifact
            )

            LocalArtifact._local_artifacts_category_to_class[
                # This should just be cls._category, but if we try to use
                # that in mypy we get:
                # TypeError: unhashable type: 'ModelPrivateAttr'
                cls._category.default  # type: ignore[attr-defined]
            ] = cls

    @classmethod
    def create_data(cls, data_dict: dict[str, Any]) -> AD:
        """Instantiate a data model from a dict."""
        return cls._data_type(**data_dict)

    @classmethod
    def _create(
        cls,
        data: dict[str, Any] | AD,
        files: Sequence[NewLocalFile | NewRemoteFile] | None = None,
    ) -> Self:
        """
        Create a new instance of this model.

        If possible, use the higher-level ``create`` method of the
        appropriate subclass instead.
        """
        if isinstance(data, dict):
            data = cls.create_data(data)
        nascent = NascentLocalArtifact()
        for file in files or ():
            match file:
                case NewLocalFile():
                    nascent.add_local_file(file)
                case NewRemoteFile():
                    nascent.add_remote_file(file)
                case _ as unreachable:
                    assert_never(unreachable)
        return cls(
            category=cls._category,
            data=data,
            files=nascent.local_files,
            remote_files=nascent.remote_files,
            content_types=nascent.content_types,
        )

    @staticmethod
    def artifact_categories() -> list[str]:
        """Return list of artifact categories."""
        return sorted(LocalArtifact._local_artifacts_category_to_class.keys())

    @pydantic.field_validator("category", mode="before")
    @classmethod
    def _validate_category(cls, category: Any) -> Any:
        """Validate that the category is known."""
        if category not in cls._local_artifacts_category_to_class:
            raise ValueError(
                f"Invalid category: '{category}'. Expected one of "
                f"{', '.join(sorted(cls._local_artifacts_category_to_class))}."
            )
        return category

    @staticmethod
    def class_from_category(category: str) -> type["LocalArtifact['Any']"]:
        """Return class sub_local_artifact."""
        category = LocalArtifact._validate_category(category)
        return LocalArtifact._local_artifacts_category_to_class[category]

    @classmethod
    def _validate_files_length(
        cls, files: dict[str, Path], number_of_files: int
    ) -> dict[str, Path]:
        """Raise ValueError if number of files is not number_of_files."""
        if (actual_number_files := len(files)) != number_of_files:
            raise ValueError(
                f"Expected number of files: {number_of_files} "
                f"Actual: {actual_number_files}"
            )
        return files

    @classmethod
    def _validate_files_end_in(
        cls, files: dict[str, Path], suffixes: Sequence[str]
    ) -> dict[str, Path]:
        """Raise ValueError if any file does not end in one of suffixes."""
        for file_name in files:
            if not file_name.endswith(tuple(suffixes)):
                raise ValueError(
                    f'Valid file suffixes: {suffixes}. '
                    f'Invalid filename: "{file_name}"'
                )
        return files

    @classmethod
    def _validate_exactly_one_file_ends_in[T: Any](
        cls, files: dict[str, T], suffix: str
    ) -> dict[str, T]:
        """Raise ValueError if files doesn't have exactly 1 file with suffix."""
        changes_files = sum(1 for file in files if file.endswith(suffix))
        if changes_files != 1:
            raise ValueError(
                f"Expecting 1 {suffix} file in {sorted(files.keys())}"
            )
        return files


class WorkRequestDebugLogs(LocalArtifact[data_models.EmptyArtifactData]):
    """
    WorkRequestDebugLogs: help debugging issues executing the task.

    Log files for debusine users in order to debug possible problems in their
    WorkRequests.
    """

    _category = data_models.ArtifactCategory.WORK_REQUEST_DEBUG_LOGS
    data: data_models.EmptyArtifactData = pydantic.Field(
        default_factory=data_models.EmptyArtifactData
    )

    @classmethod
    def create(cls, *, files: Iterable[Path]) -> Self:
        """Return a WorkRequestDebugLogs."""
        return cls._create(
            data={},
            files=[
                NewLocalFile(
                    file=file, content_type="text/plain; charset=utf-8"
                )
                for file in files
            ],
        )


class DebusineTest(LocalArtifact[data_models.EmptyArtifactData]):
    """DebusineTest: noop artifact used in Debusine tests."""

    _category = data_models.ArtifactCategory.TEST
    data: data_models.EmptyArtifactData = pydantic.Field(
        default_factory=data_models.EmptyArtifactData
    )


class PackageBuildLog(LocalArtifact[data_models.DebianPackageBuildLog]):
    """PackageBuildLog: represents a build log file."""

    _category = data_models.ArtifactCategory.PACKAGE_BUILD_LOG

    @classmethod
    def create(
        cls,
        *,
        file: Path,
        source: str,
        version: str,
        architecture: str | None = None,
        bd_uninstallable: data_models.DoseDistCheck | None = None,
    ) -> Self:
        """Return a PackageBuildLog."""
        return cls._create(
            data=data_models.DebianPackageBuildLog(
                source=source,
                version=version,
                architecture=architecture,
                filename=file.name,
                bd_uninstallable=bd_uninstallable,
            ),
            files=[
                NewLocalFile(
                    file=file, content_type="text/plain; charset=utf-8"
                )
            ],
        )

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def validate_files_length_is_one(
        cls, files: dict[str, Path]
    ) -> dict[str, Path]:
        """Validate that artifact has only one file."""
        return super()._validate_files_length(files, 1)

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def file_must_end_in_build(cls, files: dict[str, Path]) -> dict[str, Path]:
        """Raise ValueError if the file does not end in .build."""
        return super()._validate_files_end_in(files, [".build"])


@overload
def deb822dict_to_dict(
    element: dict[Any, Any] | deb822.Deb822Dict,
) -> dict[Any, Any]: ...


@overload
def deb822dict_to_dict(element: Any) -> Any: ...


def deb822dict_to_dict(element: Any) -> Any:
    """
    Traverse recursively element converting Deb822Dict to dict.

    deb822.Changes() return a Deb822Dict with some other inner-elements
    being Deb822Dict. This function traverse it and converts any elements
    being a Deb822Dict to a Python dict.

    Reason is to simplify is that json module cannot encode Deb822Dict. To
    simplify let's convert Deb822Dict as soon as possible to dict.
    """
    result: Any
    if isinstance(element, (dict, deb822.Deb822Dict)):
        result = {}
        for key, value in element.items():
            result[key] = deb822dict_to_dict(value)
    elif isinstance(element, list):
        result = []
        for item in element:
            result.append(deb822dict_to_dict(item))
    else:
        result = element

    return result


class Upload(LocalArtifact[data_models.DebianUpload]):
    """Upload: encapsulate a .changes and files listed in it."""

    _category = data_models.ArtifactCategory.UPLOAD

    @classmethod
    def create(cls, *, changes_file: Path, allow_remote: bool = False) -> Self:
        """
        Return a Upload. Add the changes_files and files listed in it.

        :param changes_file: a .changes file. Parsed by deb822.Changes.
        :param allow_remote: if True, allow files referenced by the .changes
          file to be missing locally; uploading this artifact will require
          them to already exist on the server.
        """
        data = data_models.DebianUpload(
            type="dpkg",
            changes_fields=deb822dict_to_dict(read_changes(changes_file)),
        )

        files: list[NewLocalFile | NewRemoteFile] = []
        files.append(
            NewLocalFile(
                file=changes_file, content_type="text/plain; charset=utf-8"
            )
        )

        # Add any files referenced by .changes.
        base_directory = changes_file.parent
        for file in data.changes_fields.get("Checksums-Sha256", []):
            # Checked by read_changes.
            assert is_single_path_component(file["name"])
            path = base_directory / file["name"]
            if allow_remote and not path.exists():
                files.append(
                    NewRemoteFile(
                        path_in_artifact=file["name"],
                        file_request=FileRequest(
                            size=file["size"],
                            checksums={"sha256": file["sha256"]},
                            type="file",
                        ),
                    )
                )
            else:
                files.append(NewLocalFile(file=path))

        return cls._create(data=data, files=files)

    @pydantic.model_validator(mode="after")
    def files_contain_changes(self) -> Self:
        """Raise ValueError when files does not have exactly 1 .changes file."""
        self._validate_exactly_one_file_ends_in(
            {**self.files, **self.remote_files}, ".changes"
        )
        return self

    @pydantic.model_validator(mode="after")
    def files_contains_files_in_changes(self) -> Self:
        """
        Validate that set(files) == set(files_in_changes_file).

        Exception: The .changes file must be in files but not in the .changes
        file.
        """
        files_in_meta_file_match_files(
            ".changes", deb822.Changes, {**self.files, **self.remote_files}
        )
        return self


class SourcePackage(LocalArtifact[data_models.DebianSourcePackage]):
    """SourcePackage: contains source code to be built into BinaryPackages."""

    _category = data_models.ArtifactCategory.SOURCE_PACKAGE

    @classmethod
    def create(cls, *, name: str, version: str, files: list[Path]) -> Self:
        """Return a SourcePackage setting files and data."""
        dsc_fields = {}
        for file in files:
            if file.suffix == ".dsc":
                dsc = utils.read_dsc(file)
                dsc_fields = deb822dict_to_dict(dsc)

        data = data_models.DebianSourcePackage(
            name=name,
            version=version,
            type="dpkg",
            dsc_fields=dsc_fields,
        )

        return cls._create(
            data=data, files=[NewLocalFile(file=file) for file in files]
        )

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def files_contain_one_dsc(cls, files: dict[str, Path]) -> dict[str, Path]:
        """Raise ValueError when files does not have exactly 1 .dsc file."""
        return cls._validate_exactly_one_file_ends_in(files, ".dsc")

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def files_contains_files_in_dsc(
        cls, files: dict[str, Path]
    ) -> dict[str, Path]:
        """
        Validate that set(files) == set(files_in_dsc_file).

        Exception: The .dsc file must be in files but not in the .dsc file.
        """
        files_in_meta_file_match_files(".dsc", deb822.Dsc, files)
        return files


class BinaryPackage(LocalArtifact[data_models.DebianBinaryPackage]):
    r"""BinaryPackage: encapsulates a single \*.deb / \*.udeb."""

    _category = data_models.ArtifactCategory.BINARY_PACKAGE

    @classmethod
    def create(cls, *, file: Path) -> Self:
        """Return a BinaryPackage setting file and data."""
        pkg = debfile.DebFile(file)
        try:
            control = pkg.control.debcontrol()
            control_files = sorted(
                name.lstrip("./")
                for name in pkg.control
                if name.startswith("./")
            )
        finally:
            pkg.close()
        # Work around Debian #1082838
        srcpkg_name = control.get("Source", control["Package"])
        if (m := re.match(r"^(.*) \((.*)\)$", srcpkg_name)) is not None:
            srcpkg_name, srcpkg_version = m.groups()
        else:
            srcpkg_version = control["Version"]
        data = data_models.DebianBinaryPackage(
            srcpkg_name=srcpkg_name,
            srcpkg_version=srcpkg_version,
            deb_fields=deb822dict_to_dict(control),
            deb_control_files=control_files,
        )

        return cls._create(data=data, files=[NewLocalFile(file=file)])

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def files_must_end_in_deb_or_udeb(
        cls, files: dict[str, Path]
    ) -> dict[str, Path]:
        """Raise ValueError if a file does not end in .deb or .udeb."""
        return super()._validate_files_end_in(files, [".deb", ".udeb"])

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def files_exactly_one(cls, files: dict[str, Path]) -> dict[str, Path]:
        """Raise ValueError if len(files) != 1."""
        if len(files) != 1:
            raise ValueError("Must have exactly one file")

        return files


class BinaryPackages(LocalArtifact[data_models.DebianBinaryPackages]):
    r"""BinaryPackages: encapsulates a group of \*.deb / \*.udeb."""

    _category = data_models.ArtifactCategory.BINARY_PACKAGES

    @classmethod
    def create(
        cls,
        *,
        srcpkg_name: str,
        srcpkg_version: str,
        version: str,
        architecture: str,
        files: list[Path],
    ) -> Self:
        """Return a BinaryPackages setting files and data."""
        data = data_models.DebianBinaryPackages(
            srcpkg_name=srcpkg_name,
            srcpkg_version=srcpkg_version,
            version=version,
            architecture=architecture,
            # It might be better to get this from the metadata in the
            # package, but that would be rather more effort and isn't
            # currently vital.
            packages=[path.name.split("_", 1)[0] for path in files],
        )

        return cls._create(
            data=data, files=[NewLocalFile(file=file) for file in files]
        )

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def files_must_end_in_deb_or_udeb(
        cls, files: dict[str, Path]
    ) -> dict[str, Path]:
        """Raise ValueError if a file does not end in .deb or .udeb."""
        return super()._validate_files_end_in(files, [".deb", ".udeb"])

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def files_more_than_zero(cls, files: dict[str, Path]) -> dict[str, Path]:
        """Raise ValueError if len(files) == 0."""
        if len(files) == 0:
            raise ValueError("Must have at least one file")

        return files


class LintianArtifact(LocalArtifact[data_models.DebianLintian]):
    """LintianArtifact: encapsulate result of the Lintian run."""

    _category = data_models.ArtifactCategory.LINTIAN

    @classmethod
    def create(
        cls,
        analysis: Path,
        lintian_output: Path,
        architecture: str,
        summary: data_models.DebianLintianSummary,
    ) -> Self:
        """Return a LintianArtifact with the files set."""
        data = data_models.DebianLintian(
            architecture=architecture, summary=summary
        )

        return cls._create(
            data=data,
            files=[
                NewLocalFile(file=analysis, override_name="analysis.json"),
                NewLocalFile(file=lintian_output, override_name="lintian.txt"),
            ],
        )

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def _validate_required_files(
        cls, files: dict[str, Path]
    ) -> dict[str, Path]:
        """Artifact contain "analysis.json, "lintian.txt"."""
        # The .create() method already enforces having the correct files
        # But the artifact can be created using debusine.client or a web form,
        # which do not use the .create() of LintianArtifact but the
        # LocalArtifact. That's the reason that is needed to validate
        # that the required files are attached
        required_files = {"analysis.json", "lintian.txt"}
        if files.keys() != required_files:
            raise ValueError(f"Files required: {sorted(required_files)}")

        return files

    @staticmethod
    def _file_is_json_or_raise_value_error(
        file_name: str, files: dict[str, Path]
    ) -> None:
        """Raise ValueError() if file_name in files is not valid JSON."""
        with files[file_name].open() as file:
            try:
                json.load(file)
            except JSONDecodeError as exc:
                raise ValueError(f"{file_name} is not valid JSON: {exc}")

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def _validate_file_analysis_is_json(
        cls, files: dict[str, Path]
    ) -> dict[str, Path]:
        """Validate that "analysis.json" is valid JSON."""
        cls._file_is_json_or_raise_value_error("analysis.json", files)
        return files


class AutopkgtestArtifact(LocalArtifact[data_models.DebianAutopkgtest]):
    """Autopkgtest: encapsulate result of the Autopkgtest run."""

    _category = data_models.ArtifactCategory.AUTOPKGTEST

    @classmethod
    def create(
        cls, artifact_directory: Path, data: data_models.DebianAutopkgtest
    ) -> Self:
        """Return AutopkgtestArtifact with the files and data set."""
        files = []

        for file in artifact_directory.rglob("*"):
            if not file.is_file():
                # Only add files
                continue

            if file.is_relative_to(artifact_directory / "binaries"):
                # Skip binaries/
                continue

            files.append(
                NewLocalFile(file=file, artifact_base_dir=artifact_directory)
            )

        return cls._create(data=data, files=files)


class PiupartsArtifact(LocalArtifact[data_models.DebianPiuparts]):
    """Piuparts: encapsulate result of a ``piuparts`` run."""

    _category = data_models.ArtifactCategory.PIUPARTS

    @classmethod
    def create(
        cls, *, piuparts_output: Path, data: data_models.DebianPiuparts
    ) -> Self:
        """Return a ``PiupartsArtifact`` with the files and data set."""
        return cls._create(
            data=data,
            files=[
                NewLocalFile(file=piuparts_output, override_name="piuparts.txt")
            ],
        )


class DebianSystemTarballArtifact(
    LocalArtifact[data_models.DebianSystemTarball]
):
    """
    Contain system.tar.xz file with a Debian.

    Can be used by a chroot, container, etc.
    """

    _category = data_models.ArtifactCategory.SYSTEM_TARBALL

    @classmethod
    def create(cls, tarball: Path, data: dict[str, Any]) -> Self:
        """Return a DebianSystemTarballArtifact with the tarball file."""
        data = data.copy()
        data["filename"] = tarball.name
        return cls._create(
            data=data_models.DebianSystemTarball(**data),
            files=[NewLocalFile(file=tarball)],
        )

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def _validate_file_name_ends_in_tar_xz(
        cls, files: dict[str, Path]
    ) -> dict[str, Path]:
        """Check if the artifact contains only one file and it is a .tar.xz."""
        if not len(files) == 1:
            raise ValueError(
                "DebianSystemTarballArtifact does not contain exactly one file"
            )

        if not (name := next(iter(files.keys()))).endswith(".tar.xz"):
            raise ValueError(f"Invalid file name: '{name}'. Expected .tar.xz")

        return files


class BlhcArtifact(LocalArtifact[data_models.DebianBlhc]):
    """BlhcArtifact: encapsulate result of the blhc run."""

    _category = data_models.ArtifactCategory.BLHC

    @classmethod
    def create(
        cls,
        blhc_output: Path,
        build_log: data_models.DebianPackageBuildLog | None,
    ) -> Self:
        """Return a BlhcArtifact with the files set."""
        return cls._create(
            data=data_models.DebianBlhc(
                source=None if build_log is None else build_log.source,
                version=None if build_log is None else build_log.version,
                architecture=(
                    None if build_log is None else build_log.architecture
                ),
            ),
            files=[NewLocalFile(file=blhc_output, override_name="blhc.txt")],
        )


class DebDiffArtifact(LocalArtifact[data_models.DebDiff]):
    """DebDiffArtifact: encapsulate result of the debdiff run."""

    _category = data_models.ArtifactCategory.DEBDIFF

    @classmethod
    def create(
        cls,
        debdiff_output: Path,
        original: str,
        new: str,
        *,
        label: str | None = None,
    ) -> Self:
        """Return a DebDiffArtifact with the files set."""
        return cls._create(
            data=data_models.DebDiff(original=original, new=new, label=label),
            files=[
                NewLocalFile(file=debdiff_output, override_name="debdiff.txt")
            ],
        )


class DebianSystemImageArtifact(LocalArtifact[data_models.DebianSystemImage]):
    """
    Contains a image.tar.xz file with a bootable Debian system.

    Can be used by a VM.
    """

    _category = data_models.ArtifactCategory.SYSTEM_IMAGE

    @classmethod
    def create(cls, image: Path, data: dict[str, Any]) -> Self:
        """Return a DebianSystemImageArtifact with the image file."""
        data = data.copy()
        data["filename"] = image.name
        return cls._create(
            data=data_models.DebianSystemImage(**data),
            files=[NewLocalFile(file=image)],
        )

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def _validate_files(cls, files: dict[str, Path]) -> dict[str, Path]:
        """Check if the artifact contains only one file and it's ending."""
        if not len(files) == 1:
            raise ValueError(
                "DebianSystemImageArtifact does not contain exactly one file"
            )

        name = next(iter(files.keys()))
        if not name.endswith(".tar.xz") and not name.endswith(".qcow2"):
            raise ValueError(
                f"Invalid file name: '{name}'. Expected .tar.xz or qcow2"
            )

        return files


class SigningInputArtifact(LocalArtifact[data_models.DebusineSigningInput]):
    """Input to a Sign task."""

    _category = data_models.ArtifactCategory.SIGNING_INPUT

    @classmethod
    def create(
        cls,
        files: Iterable[Path],
        base_dir: Path,
        trusted_certs: list[str] | None = None,
        binary_package_name: str | None = None,
    ) -> Self:
        """Return a new SigningInputArtifact."""
        data = data_models.DebusineSigningInput(
            trusted_certs=(
                None if trusted_certs is None else list(trusted_certs)
            ),
            binary_package_name=binary_package_name,
        )
        return cls._create(
            data=data,
            files=[
                NewLocalFile(
                    file=file,
                    artifact_base_dir=base_dir,
                    # The content-type here doesn't matter, but we don't
                    # want to involve libmagic in a potentially-sensitive
                    # signing workflow.
                    content_type="application/octet-stream",
                )
                for file in files
            ],
        )

    @pydantic.field_validator("files", mode="after")
    @classmethod
    def validate_at_least_one_file(
        cls, files: dict[str, Path]
    ) -> dict[str, Path]:
        """Validate that artifact has at least one file."""
        if not files:
            raise ValueError("Expected at least one file")
        return files


class SigningOutputArtifact(LocalArtifact[data_models.DebusineSigningOutput]):
    """Output of a Sign task."""

    _category = data_models.ArtifactCategory.SIGNING_OUTPUT

    @classmethod
    def create(
        cls,
        purpose: KeyPurpose,
        fingerprint: str,
        results: Iterable[data_models.SigningResult],
        files: Iterable[Path],
        base_dir: Path,
        binary_package_name: str | None = None,
    ) -> Self:
        """Return a new SigningOutputArtifact."""
        data = data_models.DebusineSigningOutput(
            purpose=purpose,
            fingerprint=fingerprint,
            results=list(results),
            binary_package_name=binary_package_name,
        )
        return cls._create(
            data=data,
            files=[
                NewLocalFile(
                    file=file,
                    artifact_base_dir=base_dir,
                    # The content-type here doesn't matter, but we don't
                    # want to involve libmagic in a potentially-sensitive
                    # signing workflow.
                    content_type="application/octet-stream",
                )
                for file in files
            ],
        )


class RepositoryIndex(LocalArtifact[data_models.DebianRepositoryIndex]):
    """An index file in a repository."""

    _category = data_models.ArtifactCategory.REPOSITORY_INDEX

    @classmethod
    def create(
        cls, *, file: Path, path: str, content_type: str | None = None
    ) -> Self:
        """Return a RepositoryIndex."""
        data = data_models.DebianRepositoryIndex(path=path)
        return cls._create(
            data=data,
            files=[
                NewLocalFile(
                    file=file,
                    override_name=PurePath(path).name,
                    content_type=content_type,
                )
            ],
        )


class DebugSymbols(LocalArtifact[data_models.DebianDebugSymbols]):
    """DebugSymbols: .debug ELF files extracted from a -dbgsym.deb package.

    Files are stored with the bare 40-character build-ID as the
    FileInArtifact path.
    """

    _category = data_models.ArtifactCategory.DEBUG_SYMBOLS
    data: data_models.DebianDebugSymbols

    @classmethod
    def create(cls, *, build_id_files: dict[str, Path]) -> Self:
        """Return a DebugSymbols artifact.

        :param build_id_files: mapping of 40-char build-ID to the local
          path of the corresponding .debug ELF file.
        """
        data = data_models.DebianDebugSymbols(build_ids=list(build_id_files))
        return cls._create(
            data=data,
            files=[
                NewLocalFile(
                    file=path,
                    override_name=build_id,
                    content_type="application/octet-stream",
                )
                for build_id, path in build_id_files.items()
            ],
        )

    @pydantic.model_validator(mode="after")
    def _validate_files_match_build_ids(self) -> Self:
        """Raise ValueError if files keys don't match data.build_ids."""
        errors = [
            f"missing file for build_id {build_id!r}"
            for build_id in self.data.build_ids
            if build_id not in self.files
        ]
        errors.extend(
            f"file {path!r} has no matching build_id"
            for path in self.files
            if path not in self.data.build_ids
        )
        if errors:
            raise ValueError(
                "build_ids vs files mismatch:\n" + "\n".join(errors)
            )
        return self



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used for worker pool data."""

import abc
from enum import StrEnum
from typing import Any, Literal, Self, override

import pydantic

from debusine.assets import CloudProvidersType
from debusine.tasks.models import ExtraExternalRepository


class BaseWorkerPoolDataModel(pydantic.BaseModel, abc.ABC):
    """Base pydantic model for worker pool configuration."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")


_worker_pool_specifications_models: dict[
    CloudProvidersType, type["WorkerPoolSpecifications"]
] = {}


class WorkerPoolSpecifications(BaseWorkerPoolDataModel, abc.ABC):
    """Base class for WorkerPool.specifications data models."""

    provider_type: CloudProvidersType

    @override
    def __init_subclass__(
        cls,
        provider_type: CloudProvidersType,
        **kwargs: Any,
    ) -> None:
        """Register subclass in _worker_pool_specifications_models."""
        super().__init_subclass__(**kwargs)
        _worker_pool_specifications_models[provider_type] = cls


def worker_pool_specifications_model(
    data: dict[str, Any],
) -> WorkerPoolSpecifications:
    """Load the WorkerPoolSpecifications subclass model for data."""
    model = _worker_pool_specifications_models[data["provider_type"]]
    return model.model_validate(data)


class WorkerPoolLimits(BaseWorkerPoolDataModel):
    """Specifications for limits on a WorkerPool."""

    max_active_instances: int | None = None
    target_max_seconds_per_month: int | None = None
    max_idle_seconds: int = 3600


class ScopeWorkerPoolLimits(BaseWorkerPoolDataModel):
    """Specifications for limits on a WorkerPool."""

    target_max_seconds_per_month: int | None = None
    target_latency_seconds: int | None = None


# Test Dummy:


class DummyWorkerPoolSpecification(
    WorkerPoolSpecifications, provider_type=CloudProvidersType.DUMMY
):
    """Specifications for a DummyWorkerPool."""

    provider_type: Literal[CloudProvidersType.DUMMY] = CloudProvidersType.DUMMY
    features: list[str] = []


# Amazon AWS EC2:


class DebianRelease(StrEnum):
    """Debian release that the image is based on."""

    TRIXIE = "trixie"
    FORKY = "forky"


class DebusineInstallSource(StrEnum):
    """Source for installing debusine-worker."""

    RELEASE = "release"
    BACKPORTS = "backports"
    DAILY_BUILDS = "daily-builds"
    CUSTOM = "custom"
    ENTERPRISE_STAGING = "enterprise-staging"
    ENTERPRISE = "enterprise"
    PRE_INSTALLED = "pre-installed"


class AWSEC2InstanceMarketType(StrEnum):
    """Market to purche AWS EC2 instances in."""

    ON_DEMAND = "on-demand"
    SPOT = "spot"


class MinMaxRequirements(BaseWorkerPoolDataModel):
    """Minimum and Optional Maximum Requirements."""

    Min: int
    Max: int | None = None


class AWSInstanceRequirementFilterEnum(StrEnum):
    """Values for AWSEC2InstanceRequirements options that select features."""

    INCLUDED = "included"
    EXCLUDED = "excluded"
    REQUIRED = "required"


class AWSEC2InstanceRequirements(BaseWorkerPoolDataModel):
    """AWS EC2 Instance Requirements."""

    VCpuCount: MinMaxRequirements
    MemoryMiB: MinMaxRequirements
    MemoryGiBPerVCpu: MinMaxRequirements | None = None
    SpotMaxPricePercentageOverLowestPrice: int | None = None
    MaxSpotPriceAsPercentageOfOptimalOnDemandPrice: int | None = None
    BurstablePerformance: AWSInstanceRequirementFilterEnum = (
        AWSInstanceRequirementFilterEnum.EXCLUDED
    )
    AllowedInstanceTypes: list[str] | None = None
    ExcludedInstanceTypes: list[str] | None = None


class AWSEC2NetworkInterface(BaseWorkerPoolDataModel):
    """AWS EC2 Network Interface."""

    AssociatePublicIpAddress: bool | None = None
    DeleteOnTermination: bool = True
    DeviceIndex: int = 0
    Ipv6AddressCount: int | None = None
    SubnetId: str | None = None
    Groups: list[str] | None = None

    @pydantic.field_validator("Groups")
    @classmethod
    def validate_groups(cls, value: list[str] | None) -> list[str] | None:
        """Ensure Groups is not an empty list."""
        if not value:
            return None
        return value


class AWSEC2LaunchTemplate(BaseWorkerPoolDataModel):
    """AWS EC2 Launch Template."""

    EbsOptimized: bool = False
    ImageId: str
    KeyName: str | None = None
    InstanceType: str | None = None
    InstanceRequirements: AWSEC2InstanceRequirements | None = None
    NetworkInterfaces: list[AWSEC2NetworkInterface] | None = None
    # In GiB:
    root_device_size: int | None = None
    swap_size: int | None = None
    tags: dict[str, str] = {}

    @pydantic.field_validator("NetworkInterfaces", mode="after")
    @classmethod
    def validate_network_interfaces(
        cls, value: list[AWSEC2NetworkInterface] | None
    ) -> list[AWSEC2NetworkInterface] | None:
        """Ensure NetworkInterfaces have unique indices and not empty list."""
        if value is None or len(value) == 0:
            return None
        if len({interface.DeviceIndex for interface in value}) < len(value):
            raise ValueError(
                "Each NetworkInterface's DeviceIndex must be unique"
            )
        return value

    @pydantic.model_validator(mode="after")
    def instance_type_or_requirements(self) -> Self:
        """Only one of InstanceType or InstanceRequirements is specified."""
        type_defined = self.InstanceType is not None
        req_defined = self.InstanceRequirements is not None
        if type_defined and req_defined:
            raise ValueError(
                "If InstanceType is specified, InstanceRequirements cannot be"
            )
        if not type_defined and not req_defined:
            raise ValueError(
                "InstanceType or InstanceRequirements must be specified"
            )
        return self


class AWSEC2WorkerPoolSpecification(
    WorkerPoolSpecifications, provider_type=CloudProvidersType.AWS
):
    """Specifications for an AWS EC2 WorkerPool."""

    provider_type: Literal[CloudProvidersType.AWS] = CloudProvidersType.AWS
    launch_templates: list[AWSEC2LaunchTemplate]
    instance_market_type: AWSEC2InstanceMarketType = (
        AWSEC2InstanceMarketType.SPOT
    )
    # Spot:
    max_spot_price_per_hour: float | None = None
    debian_release: DebianRelease | None = None
    debusine_install_source: DebusineInstallSource = (
        DebusineInstallSource.RELEASE
    )
    debusine_install_source_custom: ExtraExternalRepository | None = None


# Hetzner Cloud:


class HetznerCloudWorkerPoolSpecification(
    WorkerPoolSpecifications, provider_type=CloudProvidersType.HETZNER
):
    """Specifications for a Hetzner Cloud WorkerPool."""

    provider_type: Literal[CloudProvidersType.HETZNER] = (
        CloudProvidersType.HETZNER
    )
    server_type: str
    image_name: str
    ssh_keys: list[str] = []
    networks: list[str] = []
    location: str | None = None
    labels: dict[str, str] = {}
    enable_ipv4: bool = True
    enable_ipv6: bool = True
    debian_release: DebianRelease | None = None
    debusine_install_source: DebusineInstallSource = (
        DebusineInstallSource.RELEASE
    )
    debusine_install_source_custom: ExtraExternalRepository | None = None
